from ch5mpy.h5array.chunks.iter import _get_chunk_indices
from ch5mpy.indexing.slice import FullSlice, SingleIndex


def test_1d_smaller_than_nb_elements():
    assert _get_chunk_indices(10, (5,)) == ((FullSlice.whole_axis(5),),)


def test_1d_greater_than_nb_elements():
    assert _get_chunk_indices(10, (15,)) == ((FullSlice(0, 10, 1, 15),), (FullSlice(10, 15, 1, 15),))


def test_1d_greater_than_nb_elements_multiple():
    assert _get_chunk_indices(10, (30,)) == (
        (FullSlice(0, 10, 1, 30),),
        (FullSlice(10, 20, 1, 30),),
        (FullSlice(20, 30, 1, 30),),
    )


def test_2d_array_smaller_than_nb_elements():
    assert _get_chunk_indices(100, (2, 10)) == ((FullSlice.whole_axis(2), FullSlice.whole_axis(10)),)


def test_2d_array_1row():
    assert _get_chunk_indices(10, (8, 10)) == (
        (
            FullSlice.one(0, 8),
            FullSlice.whole_axis(10),
        ),
        (
            FullSlice.one(1, 8),
            FullSlice.whole_axis(10),
        ),
        (
            FullSlice.one(2, 8),
            FullSlice.whole_axis(10),
        ),
        (
            FullSlice.one(3, 8),
            FullSlice.whole_axis(10),
        ),
        (
            FullSlice.one(4, 8),
            FullSlice.whole_axis(10),
        ),
        (
            FullSlice.one(5, 8),
            FullSlice.whole_axis(10),
        ),
        (
            FullSlice.one(6, 8),
            FullSlice.whole_axis(10),
        ),
        (
            FullSlice.one(7, 8),
            FullSlice.whole_axis(10),
        ),
    )


def test_2d_array_2rows():
    assert _get_chunk_indices(20, (8, 10)) == (
        (
            FullSlice(0, 2, 1, 8),
            FullSlice.whole_axis(10),
        ),
        (
            FullSlice(2, 4, 1, 8),
            FullSlice.whole_axis(10),
        ),
        (
            FullSlice(4, 6, 1, 8),
            FullSlice.whole_axis(10),
        ),
        (
            FullSlice(6, 8, 1, 8),
            FullSlice.whole_axis(10),
        ),
    )


def test_2d_array_0rows():
    assert _get_chunk_indices(6, (8, 10)) == (
        (SingleIndex(0, 8), FullSlice(0, 6, 1, 10)),
        (SingleIndex(0, 8), FullSlice(6, 10, 1, 10)),
        (SingleIndex(1, 8), FullSlice(0, 6, 1, 10)),
        (SingleIndex(1, 8), FullSlice(6, 10, 1, 10)),
        (SingleIndex(2, 8), FullSlice(0, 6, 1, 10)),
        (SingleIndex(2, 8), FullSlice(6, 10, 1, 10)),
        (SingleIndex(3, 8), FullSlice(0, 6, 1, 10)),
        (SingleIndex(3, 8), FullSlice(6, 10, 1, 10)),
        (SingleIndex(4, 8), FullSlice(0, 6, 1, 10)),
        (SingleIndex(4, 8), FullSlice(6, 10, 1, 10)),
        (SingleIndex(5, 8), FullSlice(0, 6, 1, 10)),
        (SingleIndex(5, 8), FullSlice(6, 10, 1, 10)),
        (SingleIndex(6, 8), FullSlice(0, 6, 1, 10)),
        (SingleIndex(6, 8), FullSlice(6, 10, 1, 10)),
        (SingleIndex(7, 8), FullSlice(0, 6, 1, 10)),
        (SingleIndex(7, 8), FullSlice(6, 10, 1, 10)),
    )


def test_3d_array_smaller_than_nb_elements():
    assert _get_chunk_indices(200, (5, 5, 5)) == (
        (FullSlice.whole_axis(5), FullSlice.whole_axis(5), FullSlice.whole_axis(5)),
    )


def test_3d_array_1_array():
    assert _get_chunk_indices(30, (5, 5, 5)) == (
        (FullSlice.one(0, 5), FullSlice.whole_axis(5), FullSlice.whole_axis(5)),
        (FullSlice.one(1, 5), FullSlice.whole_axis(5), FullSlice.whole_axis(5)),
        (FullSlice.one(2, 5), FullSlice.whole_axis(5), FullSlice.whole_axis(5)),
        (FullSlice.one(3, 5), FullSlice.whole_axis(5), FullSlice.whole_axis(5)),
        (FullSlice.one(4, 5), FullSlice.whole_axis(5), FullSlice.whole_axis(5)),
    )


def test_3d_array_2_arrays():
    assert _get_chunk_indices(60, (5, 5, 5)) == (
        (FullSlice(0, 2, 1, 5), FullSlice.whole_axis(5), FullSlice.whole_axis(5)),
        (FullSlice(2, 4, 1, 5), FullSlice.whole_axis(5), FullSlice.whole_axis(5)),
        (FullSlice(4, 5, 1, 5), FullSlice.whole_axis(5), FullSlice.whole_axis(5)),
    )


def test_3d_array_2rows():
    assert _get_chunk_indices(20, (5, 5, 5)) == (
        (SingleIndex(0, 5), FullSlice(0, 4, 1, 5), FullSlice.whole_axis(5)),
        (SingleIndex(0, 5), FullSlice(4, 5, 1, 5), FullSlice.whole_axis(5)),
        (SingleIndex(1, 5), FullSlice(0, 4, 1, 5), FullSlice.whole_axis(5)),
        (SingleIndex(1, 5), FullSlice(4, 5, 1, 5), FullSlice.whole_axis(5)),
        (SingleIndex(2, 5), FullSlice(0, 4, 1, 5), FullSlice.whole_axis(5)),
        (SingleIndex(2, 5), FullSlice(4, 5, 1, 5), FullSlice.whole_axis(5)),
        (SingleIndex(3, 5), FullSlice(0, 4, 1, 5), FullSlice.whole_axis(5)),
        (SingleIndex(3, 5), FullSlice(4, 5, 1, 5), FullSlice.whole_axis(5)),
        (SingleIndex(4, 5), FullSlice(0, 4, 1, 5), FullSlice.whole_axis(5)),
        (SingleIndex(4, 5), FullSlice(4, 5, 1, 5), FullSlice.whole_axis(5)),
    )


def test_3d_array_1row():
    assert _get_chunk_indices(6, (5, 5, 5)) == (
        (SingleIndex(0, 5), FullSlice.one(0, 5), FullSlice.whole_axis(5)),
        (SingleIndex(0, 5), FullSlice.one(1, 5), FullSlice.whole_axis(5)),
        (SingleIndex(0, 5), FullSlice.one(2, 5), FullSlice.whole_axis(5)),
        (SingleIndex(0, 5), FullSlice.one(3, 5), FullSlice.whole_axis(5)),
        (SingleIndex(0, 5), FullSlice.one(4, 5), FullSlice.whole_axis(5)),
        (SingleIndex(1, 5), FullSlice.one(0, 5), FullSlice.whole_axis(5)),
        (SingleIndex(1, 5), FullSlice.one(1, 5), FullSlice.whole_axis(5)),
        (SingleIndex(1, 5), FullSlice.one(2, 5), FullSlice.whole_axis(5)),
        (SingleIndex(1, 5), FullSlice.one(3, 5), FullSlice.whole_axis(5)),
        (SingleIndex(1, 5), FullSlice.one(4, 5), FullSlice.whole_axis(5)),
        (SingleIndex(2, 5), FullSlice.one(0, 5), FullSlice.whole_axis(5)),
        (SingleIndex(2, 5), FullSlice.one(1, 5), FullSlice.whole_axis(5)),
        (SingleIndex(2, 5), FullSlice.one(2, 5), FullSlice.whole_axis(5)),
        (SingleIndex(2, 5), FullSlice.one(3, 5), FullSlice.whole_axis(5)),
        (SingleIndex(2, 5), FullSlice.one(4, 5), FullSlice.whole_axis(5)),
        (SingleIndex(3, 5), FullSlice.one(0, 5), FullSlice.whole_axis(5)),
        (SingleIndex(3, 5), FullSlice.one(1, 5), FullSlice.whole_axis(5)),
        (SingleIndex(3, 5), FullSlice.one(2, 5), FullSlice.whole_axis(5)),
        (SingleIndex(3, 5), FullSlice.one(3, 5), FullSlice.whole_axis(5)),
        (SingleIndex(3, 5), FullSlice.one(4, 5), FullSlice.whole_axis(5)),
        (SingleIndex(4, 5), FullSlice.one(0, 5), FullSlice.whole_axis(5)),
        (SingleIndex(4, 5), FullSlice.one(1, 5), FullSlice.whole_axis(5)),
        (SingleIndex(4, 5), FullSlice.one(2, 5), FullSlice.whole_axis(5)),
        (SingleIndex(4, 5), FullSlice.one(3, 5), FullSlice.whole_axis(5)),
        (SingleIndex(4, 5), FullSlice.one(4, 5), FullSlice.whole_axis(5)),
    )


def test_3d_array_0rows():
    assert _get_chunk_indices(3, (5, 5, 5)) == (
        (SingleIndex(0, 5), SingleIndex(0, 5), FullSlice(0, 3, 1, 5)),
        (SingleIndex(0, 5), SingleIndex(0, 5), FullSlice(3, 5, 1, 5)),
        (SingleIndex(0, 5), SingleIndex(1, 5), FullSlice(0, 3, 1, 5)),
        (SingleIndex(0, 5), SingleIndex(1, 5), FullSlice(3, 5, 1, 5)),
        (SingleIndex(0, 5), SingleIndex(2, 5), FullSlice(0, 3, 1, 5)),
        (SingleIndex(0, 5), SingleIndex(2, 5), FullSlice(3, 5, 1, 5)),
        (SingleIndex(0, 5), SingleIndex(3, 5), FullSlice(0, 3, 1, 5)),
        (SingleIndex(0, 5), SingleIndex(3, 5), FullSlice(3, 5, 1, 5)),
        (SingleIndex(0, 5), SingleIndex(4, 5), FullSlice(0, 3, 1, 5)),
        (SingleIndex(0, 5), SingleIndex(4, 5), FullSlice(3, 5, 1, 5)),
        (SingleIndex(1, 5), SingleIndex(0, 5), FullSlice(0, 3, 1, 5)),
        (SingleIndex(1, 5), SingleIndex(0, 5), FullSlice(3, 5, 1, 5)),
        (SingleIndex(1, 5), SingleIndex(1, 5), FullSlice(0, 3, 1, 5)),
        (SingleIndex(1, 5), SingleIndex(1, 5), FullSlice(3, 5, 1, 5)),
        (SingleIndex(1, 5), SingleIndex(2, 5), FullSlice(0, 3, 1, 5)),
        (SingleIndex(1, 5), SingleIndex(2, 5), FullSlice(3, 5, 1, 5)),
        (SingleIndex(1, 5), SingleIndex(3, 5), FullSlice(0, 3, 1, 5)),
        (SingleIndex(1, 5), SingleIndex(3, 5), FullSlice(3, 5, 1, 5)),
        (SingleIndex(1, 5), SingleIndex(4, 5), FullSlice(0, 3, 1, 5)),
        (SingleIndex(1, 5), SingleIndex(4, 5), FullSlice(3, 5, 1, 5)),
        (SingleIndex(2, 5), SingleIndex(0, 5), FullSlice(0, 3, 1, 5)),
        (SingleIndex(2, 5), SingleIndex(0, 5), FullSlice(3, 5, 1, 5)),
        (SingleIndex(2, 5), SingleIndex(1, 5), FullSlice(0, 3, 1, 5)),
        (SingleIndex(2, 5), SingleIndex(1, 5), FullSlice(3, 5, 1, 5)),
        (SingleIndex(2, 5), SingleIndex(2, 5), FullSlice(0, 3, 1, 5)),
        (SingleIndex(2, 5), SingleIndex(2, 5), FullSlice(3, 5, 1, 5)),
        (SingleIndex(2, 5), SingleIndex(3, 5), FullSlice(0, 3, 1, 5)),
        (SingleIndex(2, 5), SingleIndex(3, 5), FullSlice(3, 5, 1, 5)),
        (SingleIndex(2, 5), SingleIndex(4, 5), FullSlice(0, 3, 1, 5)),
        (SingleIndex(2, 5), SingleIndex(4, 5), FullSlice(3, 5, 1, 5)),
        (SingleIndex(3, 5), SingleIndex(0, 5), FullSlice(0, 3, 1, 5)),
        (SingleIndex(3, 5), SingleIndex(0, 5), FullSlice(3, 5, 1, 5)),
        (SingleIndex(3, 5), SingleIndex(1, 5), FullSlice(0, 3, 1, 5)),
        (SingleIndex(3, 5), SingleIndex(1, 5), FullSlice(3, 5, 1, 5)),
        (SingleIndex(3, 5), SingleIndex(2, 5), FullSlice(0, 3, 1, 5)),
        (SingleIndex(3, 5), SingleIndex(2, 5), FullSlice(3, 5, 1, 5)),
        (SingleIndex(3, 5), SingleIndex(3, 5), FullSlice(0, 3, 1, 5)),
        (SingleIndex(3, 5), SingleIndex(3, 5), FullSlice(3, 5, 1, 5)),
        (SingleIndex(3, 5), SingleIndex(4, 5), FullSlice(0, 3, 1, 5)),
        (SingleIndex(3, 5), SingleIndex(4, 5), FullSlice(3, 5, 1, 5)),
        (SingleIndex(4, 5), SingleIndex(0, 5), FullSlice(0, 3, 1, 5)),
        (SingleIndex(4, 5), SingleIndex(0, 5), FullSlice(3, 5, 1, 5)),
        (SingleIndex(4, 5), SingleIndex(1, 5), FullSlice(0, 3, 1, 5)),
        (SingleIndex(4, 5), SingleIndex(1, 5), FullSlice(3, 5, 1, 5)),
        (SingleIndex(4, 5), SingleIndex(2, 5), FullSlice(0, 3, 1, 5)),
        (SingleIndex(4, 5), SingleIndex(2, 5), FullSlice(3, 5, 1, 5)),
        (SingleIndex(4, 5), SingleIndex(3, 5), FullSlice(0, 3, 1, 5)),
        (SingleIndex(4, 5), SingleIndex(3, 5), FullSlice(3, 5, 1, 5)),
        (SingleIndex(4, 5), SingleIndex(4, 5), FullSlice(0, 3, 1, 5)),
        (SingleIndex(4, 5), SingleIndex(4, 5), FullSlice(3, 5, 1, 5)),
    )


def test_3d_array():
    assert _get_chunk_indices(3, (3, 4, 5)) == (
        (SingleIndex(0, 3), SingleIndex(0, 4), FullSlice(0, 3, 1, 5)),
        (SingleIndex(0, 3), SingleIndex(0, 4), FullSlice(3, 5, 1, 5)),
        (SingleIndex(0, 3), SingleIndex(1, 4), FullSlice(0, 3, 1, 5)),
        (SingleIndex(0, 3), SingleIndex(1, 4), FullSlice(3, 5, 1, 5)),
        (SingleIndex(0, 3), SingleIndex(2, 4), FullSlice(0, 3, 1, 5)),
        (SingleIndex(0, 3), SingleIndex(2, 4), FullSlice(3, 5, 1, 5)),
        (SingleIndex(0, 3), SingleIndex(3, 4), FullSlice(0, 3, 1, 5)),
        (SingleIndex(0, 3), SingleIndex(3, 4), FullSlice(3, 5, 1, 5)),
        (SingleIndex(1, 3), SingleIndex(0, 4), FullSlice(0, 3, 1, 5)),
        (SingleIndex(1, 3), SingleIndex(0, 4), FullSlice(3, 5, 1, 5)),
        (SingleIndex(1, 3), SingleIndex(1, 4), FullSlice(0, 3, 1, 5)),
        (SingleIndex(1, 3), SingleIndex(1, 4), FullSlice(3, 5, 1, 5)),
        (SingleIndex(1, 3), SingleIndex(2, 4), FullSlice(0, 3, 1, 5)),
        (SingleIndex(1, 3), SingleIndex(2, 4), FullSlice(3, 5, 1, 5)),
        (SingleIndex(1, 3), SingleIndex(3, 4), FullSlice(0, 3, 1, 5)),
        (SingleIndex(1, 3), SingleIndex(3, 4), FullSlice(3, 5, 1, 5)),
        (SingleIndex(2, 3), SingleIndex(0, 4), FullSlice(0, 3, 1, 5)),
        (SingleIndex(2, 3), SingleIndex(0, 4), FullSlice(3, 5, 1, 5)),
        (SingleIndex(2, 3), SingleIndex(1, 4), FullSlice(0, 3, 1, 5)),
        (SingleIndex(2, 3), SingleIndex(1, 4), FullSlice(3, 5, 1, 5)),
        (SingleIndex(2, 3), SingleIndex(2, 4), FullSlice(0, 3, 1, 5)),
        (SingleIndex(2, 3), SingleIndex(2, 4), FullSlice(3, 5, 1, 5)),
        (SingleIndex(2, 3), SingleIndex(3, 4), FullSlice(0, 3, 1, 5)),
        (SingleIndex(2, 3), SingleIndex(3, 4), FullSlice(3, 5, 1, 5)),
    )



import pytest
from pathlib import Path

from ch5mpy import File
from ch5mpy import H5Mode
from ch5mpy import H5Array
from ch5mpy import write_object


class ComplexObject:
    def __init__(self, value: int):
        self.value = value

    def __repr__(self) -> str:
        return f"CO({self.value})"


@pytest.fixture
def co_array() -> H5Array:
    data = [1, 2, 3, 4, 5, 6, 7]

    with File("h5_str_array", H5Mode.WRITE_TRUNCATE) as h5_file:
        write_object(h5_file, "data", data)

    yield H5Array(File("h5_str_array", H5Mode.READ_WRITE)["data"]).maptype(ComplexObject)

    Path("h5_str_array").unlink()


def test_co_array_dtype(co_array):
    assert co_array.dtype == object


def test_co_array_equals(co_array):
    first_element = co_array[0]
    assert isinstance(first_element, ComplexObject) and first_element.value == 1


def test_co_array_should_repr_str(co_array):
    assert repr(co_array) == "H5Array([CO(1), CO(2), CO(3), ..., CO(5), CO(6), CO(7)], shape=(7,), dtype=object)"



import numpy as np
import pytest


def test_str_array_dtype(str_array):
    assert str_array.dtype == np.dtype("<U3")


def test_str_array_equals(str_array):
    assert np.array_equal(str_array, ["a", "bc", "d", "efg", "h"])


def test_str_array_should_convert_to_numpy_array(str_array):
    np_arr = np.array(str_array)
    assert type(np_arr) == np.ndarray


def test_str_array_repr(str_array):
    assert repr(str_array) == "H5Array(['a', 'bc', 'd', 'efg', 'h'], shape=(5,), dtype='<U3')"


def test_setitem_str(str_array):
    str_array[[0, 1]] = np.array(["A", "BBBB"])
    assert np.array_equal(str_array, ["A", "BBBB", "d", "efg", "h"])
    assert str_array.dtype == np.dtype("<U4")


def test_array_str_type_casting(str_array, array):
    assert str_array.astype(str).dtype == np.dtype("<U3")
    assert array.astype(str).dtype == np.dtype("<U32")


@pytest.mark.xfail
def test_array_str_cast_int_should_fail(str_array):
    _ = str_array.astype(int)[:]


def test_iter_chunks_str_array(str_array):
    _, chunk = list(str_array.iter_chunks())[0]
    assert np.issubdtype(chunk.dtype, str)


def test_str_array_should_convert_to_numpy(str_array):
    assert type(np.array(str_array)) == np.ndarray
    assert type(np.array(str_array[0:1])) == np.ndarray



import numpy as np


def test_should_add_inplace(chunked_array):
    chunked_array += 1
    assert np.array_equal(chunked_array, np.arange(100).reshape((10, 10)) + 1)


def test_should_resize_larger(chunked_array):
    chunked_array.expand(2)

    assert chunked_array.shape == (12, 10)


def test_should_resize_smaller(chunked_array):
    chunked_array.contract(1, 1)

    assert chunked_array.shape == (10, 9)



from __future__ import annotations

from pathlib import Path
from typing import Generator

import numpy as np
import pytest

from ch5mpy import File, H5Array, H5Dict, H5Mode, write_object


@pytest.fixture
def backed_dict() -> Generator[H5Dict, None, None]:
    data = {
        "a": 1,
        "b": [1, 2, 3],
        "c": {"d": "test", "e": np.arange(100)},
        "f": np.zeros((10, 10, 10)),
    }

    with File("backed_dict", H5Mode.WRITE_TRUNCATE) as h5_file:
        write_object(h5_file, "uns", data)

    yield H5Dict(File("backed_dict", H5Mode.READ_WRITE)["uns"])

    Path("backed_dict").unlink()


def test_backed_dict_creation(backed_dict):
    assert isinstance(backed_dict, H5Dict)


def test_backed_dict_can_iterate_through_keys(backed_dict):
    assert list(iter(backed_dict)) == ["a", "b", "c", "f"]


def test_backed_dict_has_correct_keys(backed_dict):
    assert list(backed_dict.keys()) == ["a", "b", "c", "f"]


def test_backed_dict_can_get_regular_values(backed_dict):
    assert backed_dict["a"] == 1


def test_backed_dict_can_get_string(backed_dict):
    assert backed_dict["c"]["d"] == "test"


def test_backed_dict_can_get_with_getattr(backed_dict):
    assert isinstance(backed_dict.c.e, H5Array)


def test_backed_dict_should_return_string(backed_dict):
    assert isinstance(backed_dict["c"]["d"], str)


def test_backed_dict_gets_nested_backed_dicts(backed_dict):
    assert isinstance(backed_dict["c"], H5Dict)


def test_backed_dict_has_correct_values(backed_dict):
    values_list = list(backed_dict.values())

    assert (
        values_list[0] == 1 and np.array_equal(values_list[1], [1, 2, 3]) and list(values_list[2].keys()) == ["d", "e"]
    )


def test_backed_dict_has_correct_items(backed_dict):
    assert list(backed_dict.items())[0] == ("a", 1)


def test_backed_dict_can_set_regular_value(backed_dict):
    backed_dict["a"] = 5

    assert backed_dict["a"] == 5


def test_backed_dict_can_set_array_value(backed_dict):
    backed_dict["b"][1] = 6

    assert np.all(backed_dict["b"] == [1, 6, 3])


def test_backed_dict_can_set_new_regular_value(backed_dict):
    backed_dict["x"] = 9

    assert backed_dict["x"] == 9


def test_backed_dict_can_set_new_array(backed_dict):
    backed_dict["y"] = np.array([1, 2, 3])

    assert np.all(backed_dict["y"] == [1, 2, 3])


def test_backed_dict_can_set_new_dict(backed_dict):
    backed_dict["z"] = {"l": 10, "m": [10, 11, 12], "n": {"o": 13}}

    assert (
        isinstance(backed_dict["z"], H5Dict)
        and backed_dict["z"]["l"] == 10
        and np.all(backed_dict["z"]["m"] == [10, 11, 12])
        and isinstance(backed_dict["z"]["n"], H5Dict)
        and backed_dict["z"]["n"]["o"] == 13
    )


def test_backed_dict_can_delete_regular_value(backed_dict):
    del backed_dict["a"]

    assert "a" not in backed_dict.keys()


def test_backed_dict_can_delete_array(backed_dict):
    del backed_dict["b"]

    assert "b" not in backed_dict.keys()


def test_backed_dict_can_delete_dict(backed_dict):
    del backed_dict["c"]

    assert "c" not in backed_dict.keys()


def test_backed_dict_can_close_file(backed_dict):
    backed_dict.close()

    with pytest.raises(ValueError):
        _ = backed_dict["x"]


def test_backed_dict_copy_should_be_regular_dict(backed_dict):
    c = backed_dict.copy()

    assert isinstance(c, dict)


def test_backed_dict_copy_should_have_same_keys(backed_dict):
    c = backed_dict.copy()

    assert c.keys() == backed_dict.keys()


def test_backed_dict_copy_nested_backed_dict_should_be_dict(backed_dict):
    c = backed_dict.copy()

    assert isinstance(c["c"], dict)


def test_backed_dict_copy_dataset_proxy_should_be_array(backed_dict):
    c = backed_dict.copy()

    assert type(c["b"]) == np.ndarray


class ComplexObject:
    def __init__(self, value: int):
        self.value = value

    def __repr__(self) -> str:
        return f"CO({self.value})"

    def __h5_write__(self, values: H5Dict) -> None:
        values.attributes["value"] = self.value

    @classmethod
    def __h5_read__(cls, values: H5Dict) -> ComplexObject:
        return ComplexObject(values.attributes["value"])


def test_backed_dict_can_store_complex_objects(backed_dict):
    backed_dict["g"] = {"a": ComplexObject(1), "b": ComplexObject(2)}

    assert isinstance(backed_dict["g"]["a"], ComplexObject)



import numpy as np

import ch5mpy as ch


def test_ones(group):
    arr = ch.ones(5, "test", group)

    assert isinstance(arr, ch.H5Array)
    assert arr.shape == (5,)
    assert np.array_equal(arr, np.ones(5))


def test_ones_set_in_H5Dict(group):
    d = ch.H5Dict(group)

    d["test_set"] = ch.ones.anonymous((5, 5))

    assert np.array_equal(d["test_set"], np.ones((5, 5)))


def test_ones_set_in_H5Dict_nested(group):
    d = ch.H5Dict(group)

    d["test_set_nested"] = {"some_a": {"some_b": ch.ones.anonymous((5, 5))}}

    assert np.array_equal(d["test_set_nested"]["some_a"]["some_b"], np.ones((5, 5)))



from pathlib import Path
from typing import Generator

import numpy as np
import pytest

import ch5mpy
from ch5mpy import File, H5Array, H5Mode, write_object


def test_array_equal(small_array):
    assert np.array_equal(small_array, [1, 2, 3, 4, 5])


def test_sum(array):
    assert np.sum(array) == 4950


def test_sum_with_initial_value(array):
    assert np.sum(array, initial=1) == 4951


def test_sum_with_output(array):
    out = np.array(0)
    assert np.sum(array, out=out) == 4950


def test_sum_with_initialized_output(array):
    out = np.array(10)
    assert np.sum(array, out=out) == 4950


def test_sum_with_output_and_diff_dtype(array):
    out = np.array(0, dtype=float)
    array += 0.5
    assert np.sum(array, out=out, dtype=int) == 4950


def test_sum_keepdims(array):
    assert np.array_equal(np.sum(array, keepdims=True), [[4950]])


def test_sum_with_axis(array):
    assert np.array_equal(np.sum(array, axis=0), [450, 460, 470, 480, 490, 500, 510, 520, 530, 540])


def test_sum_axis_0(small_large_array):
    with ch5mpy.options(max_memory=str(3 * small_large_array.dtype.itemsize)):
        assert np.array_equal(
            np.sum(small_large_array, axis=0),
            np.array([[60, 63, 66, 69, 72], [75, 78, 81, 84, 87], [90, 93, 96, 99, 102], [105, 108, 111, 114, 117]]),
        )


def test_sum_axis_2(small_large_array):
    with ch5mpy.options(max_memory=str(3 * small_large_array.dtype.itemsize)):
        assert np.array_equal(
            np.sum(small_large_array, axis=2),
            np.array([[10.0, 35.0, 60.0, 85.0], [110.0, 135.0, 160.0, 185.0], [210.0, 235.0, 260.0, 285.0]]),
        )


def test_sum_where(small_large_array):
    assert np.array_equal(
        np.sum(small_large_array, axis=1, where=[True, False, False, True, True]),
        np.array([[30.0, 0.0, 0.0, 42.0, 46.0], [110.0, 0.0, 0.0, 122.0, 126.0], [190.0, 0.0, 0.0, 202.0, 206.0]]),
    )


def test_sum_where_multiple_rows(small_large_array):
    with ch5mpy.options(max_memory=str(40 * small_large_array.dtype.itemsize)):
        assert np.array_equal(
            np.sum(small_large_array, axis=1, where=[True, False, False, True, True]),
            np.array([[30.0, 0.0, 0.0, 42.0, 46.0], [110.0, 0.0, 0.0, 122.0, 126.0], [190.0, 0.0, 0.0, 202.0, 206.0]]),
        )


def test_sum_where_few_elements(small_large_array):
    with ch5mpy.options(max_memory=str(3 * small_large_array.dtype.itemsize)):
        assert np.array_equal(
            np.sum(small_large_array, axis=1, where=[True, False, False, True, True]),
            np.array([[30.0, 0.0, 0.0, 42.0, 46.0], [110.0, 0.0, 0.0, 122.0, 126.0], [190.0, 0.0, 0.0, 202.0, 206.0]]),
        )


def test_all(array):
    assert not np.all(array)


def test_all_axis_keepdims(array):
    assert np.array_equal(
        np.all(array, axis=1, keepdims=True),
        np.array([[False], [True], [True], [True], [True], [True], [True], [True], [True], [True]]),
    )


def test_floor(array):
    array += 0.7
    assert np.array_equal(np.floor(array), np.arange(100.0).reshape((10, 10)))


def test_ceil(array):
    array += 0.2
    assert np.array_equal(np.ceil(array), np.arange(1.0, 101.0).reshape((10, 10)))


# def test_ceil_large(large_array):
#     np.ceil(large_array)


def test_trunc(array):
    array += 0.2
    assert np.array_equal(np.trunc(array), np.arange(100.0).reshape((10, 10)))


def test_prod(array):
    assert np.array_equal(
        np.prod(array, axis=0, where=[True, False, False, False, False, True, True, True, True, True]),
        [0, 1, 1, 1, 1, 6393838623046875, 9585618768101376, 13865696119905399, 19511273389031424, 26853950884211451],
    )


def test_exp(array):
    assert np.array_equal(np.exp(array), np.exp(np.arange(100.0).reshape((10, 10))))


def test_exp_where(small_array):
    out = np.array([1.0, 1.0, 1.0, 1.0, 1.0])
    np.exp(small_array, out, where=[True, False, False, True, True])

    assert np.array_equal(out, [np.exp(1), 1.0, 1.0, np.exp(4), np.exp(5)])


def test_expm1(small_array):
    assert np.array_equal(np.expm1(small_array), np.expm1([1, 2, 3, 4, 5]))


def test_isfinite(small_array):
    small_array[0] = np.inf
    assert np.array_equal(np.isfinite(small_array), [False, True, True, True, True])


def test_equal(small_array):
    assert np.all(np.equal(small_array, [1, 2, 3, 4, 5]))
    assert np.array_equal(np.equal(small_array, 2), [False, True, False, False, False])


def test_equal_other_shape(small_array):
    assert np.array_equal(
        small_array == np.array([1, 2, 3, 4, 5]).reshape((-1, 1)),
        np.array(
            [
                [True, False, False, False, False],
                [False, True, False, False, False],
                [False, False, True, False, False],
                [False, False, False, True, False],
                [False, False, False, False, True],
            ]
        ),
    )


def test_equal_other_shape_str(str_array):
    assert np.array_equal(
        str_array == np.array(["a", "bc", "d", "efg", "h"]).reshape((-1, 1)),
        np.array(
            [
                [True, False, False, False, False],
                [False, True, False, False, False],
                [False, False, True, False, False],
                [False, False, False, True, False],
                [False, False, False, False, True],
            ]
        ),
    )


def test_equal_view(array):
    subarr = array[[[0], [1], [2]], [0, 1]]
    assert np.all(subarr == np.array([[0, 1], [10, 11], [20, 21]]))


def test_equal_view_other_shape(small_array):
    subarr = small_array[[3, 2, 4]]
    assert np.array_equal(
        subarr == np.array([3, 5, 4]).reshape((-1, 1)),
        np.array([[False, True, False], [False, False, True], [True, False, False]]),
    )


def test_equal_2d_array(array):
    assert np.array_equal(array[[[0], [1], [2]], 0], np.array([[0.0], [10.0], [20.0]]))


def test_equal_where(small_array):
    assert np.array_equal(
        np.equal(small_array, [1, 2, 3, 3, 3], where=[True, True, False, False, True]),
        [True, True, False, False, False],
    )


def test_equal_with_broadcast():
    data = np.ones((10, 2))

    with File("h5_equal_array", H5Mode.WRITE_TRUNCATE) as h5_file:
        write_object(h5_file, "data", data)

    arr = H5Array(File("h5_equal_array", H5Mode.READ_WRITE)["data"])

    # -----------------------------------------------------
    assert np.all(arr == [1, 1])

    # -----------------------------------------------------
    Path("h5_equal_array").unlink()


def test_greater_element(small_array):
    assert np.array_equal(np.greater(small_array, 3), [False, False, False, True, True])


def test_multiply(small_array):
    assert np.array_equal(np.multiply(small_array, [2, 3, 4, 5, 6]), [2, 6, 12, 20, 30])


def test_multiply_where(small_array):
    assert np.array_equal(
        np.multiply(small_array, [2, 3, 4, 5, 6], where=[True, False, False, True, True]), [2, 2, 3, 20, 30]
    )


def test_isinf(small_array):
    small_array[1] = np.inf
    assert np.array_equal(np.isinf(small_array), [False, True, False, False, False])


def test_isinf_out(small_array):
    small_array[3] = np.inf
    out = np.array([1, 2, 3, 4, 5])
    np.isinf(small_array, where=[True, False, False, True, True], out=out)
    assert np.array_equal(out, [False, 2, 3, True, False])


def test_logical_and(small_array):
    assert np.array_equal(
        np.logical_and(small_array, [True, True, False, False, True]), [True, True, False, False, True]
    )


@pytest.fixture
def repeating_array() -> Generator[H5Array, None, None]:
    data = np.array([[1.0, 2, 1, 1, 2, 1], [2, 1, 0, 1, 1, 2]])

    with File("h5_r_array", H5Mode.WRITE_TRUNCATE) as h5_file:
        write_object(h5_file, "data", data)

    yield H5Array(File("h5_r_array", H5Mode.READ_WRITE)["data"])

    Path("h5_r_array").unlink()


def test_unique(repeating_array):
    assert np.array_equal(np.unique(repeating_array), [0, 1, 2])


def test_unique_by_chunks(repeating_array):
    with ch5mpy.options(max_memory=3 * repeating_array.dtype.itemsize):
        assert np.array_equal(np.unique(repeating_array), [0, 1, 2])


def test_unique_with_index(repeating_array):
    with ch5mpy.options(max_memory=3 * repeating_array.dtype.itemsize):
        unique, index = np.unique(repeating_array, return_index=True)
        assert np.array_equal(unique, [0, 1, 2])
        assert np.array_equal(index, [8, 0, 1])


def test_unique_with_counts(repeating_array):
    with ch5mpy.options(max_memory=3 * repeating_array.dtype.itemsize):
        _, counts = np.unique(repeating_array, return_counts=True)
        assert np.array_equal(counts, [1, 7, 4])


def test_unique_not_equal_nan(repeating_array):
    repeating_array[0, 0] = np.nan
    repeating_array[1, 4] = np.nan
    unique = np.unique(repeating_array, equal_nan=False)
    assert np.array_equal(unique, [0, 1, 2, np.nan, np.nan], equal_nan=True)


def test_unique_not_equal_nan_with_counts(repeating_array):
    repeating_array[0, 0] = np.nan
    repeating_array[1, 4] = np.nan
    _, counts = np.unique(repeating_array, return_counts=True, equal_nan=False)
    assert np.array_equal(counts, [1, 5, 4, 1, 1])


def test_in1d_np_in_h5(small_array):
    with ch5mpy.options(max_memory=3 * small_array.dtype.itemsize):
        assert np.array_equal(np.in1d([4, 1, 7], small_array), [True, True, False])


def test_in1d_np_in_h5_invert(small_array):
    with ch5mpy.options(max_memory=3 * small_array.dtype.itemsize):
        assert np.array_equal(np.in1d([4, 1, 7], small_array, invert=True), [False, False, True])


def test_in1d_h5_in_np(small_array):
    with ch5mpy.options(max_memory=3 * small_array.dtype.itemsize):
        assert np.array_equal(np.in1d(small_array, [4, 1, 7]), [True, False, False, True, False])


def test_in1d_h5_in_h5(small_array):
    with ch5mpy.options(max_memory=3 * small_array.dtype.itemsize):
        assert np.array_equal(np.in1d(small_array, small_array), [True, True, True, True, True])


def test_isin_h5_in_np(small_array):
    with ch5mpy.options(max_memory=3 * small_array.dtype.itemsize):
        res = np.isin(small_array, [4, 1, 7])
        assert res.shape == (5,)
        assert np.array_equal(res, [True, False, False, True, False])


def test_isin_np_in_h5(small_array):
    with ch5mpy.options(max_memory=3 * small_array.dtype.itemsize):
        res = np.isin([4, 1, 7], small_array)
        assert res.shape == (3,)
        assert np.array_equal(res, [True, True, False])


def test_isin_h5_in_h5(array, small_array):
    with ch5mpy.options(max_memory=3 * array.dtype.itemsize):
        res = np.isin(array, small_array)
        assert res.shape == (10, 10)

        expected = np.zeros((10, 10), dtype=bool)
        expected[0, [1, 2, 3, 4, 5]] = True
        assert np.array_equal(res, expected)


def test_amax(array):
    assert np.amax(array) == 99
    assert np.array_equal(np.amax(array, axis=1), np.array([9, 19, 29, 39, 49, 59, 69, 79, 89, 99]))


def test_hstack(array):
    assert np.hstack((array, ch5mpy.arange_nd((10, 10)))).shape == (10, 20)


def test_mean(array):
    assert np.mean(array) == 49.5


def test_mean_on_columns(array):
    assert np.array_equal(np.mean(array, axis=1), [4.5, 14.5, 24.5, 34.5, 44.5, 54.5, 64.5, 74.5, 84.5, 94.5])


def test_insert_1D(small_array):
    np.insert(small_array, 2, 0)
    assert np.array_equal(small_array, [1, 2, 0, 3, 4, 5])


def test_insert_1D_start(small_array):
    np.insert(small_array, 0, 0)
    assert np.array_equal(small_array, [0, 1, 2, 3, 4, 5])


def test_insert_1D_end(small_array):
    np.insert(small_array, 5, 0)
    assert np.array_equal(small_array, [1, 2, 3, 4, 5, 0])


def test_insert_1D_negative(small_array):
    np.insert(small_array, -1, 0)
    assert np.array_equal(small_array, [1, 2, 3, 4, 0, 5])


def test_insert_3D(small_large_array):
    np.insert(small_large_array, 1, -1, axis=1)
    assert np.array_equal(
        small_large_array,
        [
            [[0, 1, 2, 3, 4], [-1, -1, -1, -1, -1], [5, 6, 7, 8, 9], [10, 11, 12, 13, 14], [15, 16, 17, 18, 19]],
            [
                [20, 21, 22, 23, 24],
                [-1, -1, -1, -1, -1],
                [25, 26, 27, 28, 29],
                [30, 31, 32, 33, 34],
                [35, 36, 37, 38, 39],
            ],
            [
                [40, 41, 42, 43, 44],
                [-1, -1, -1, -1, -1],
                [45, 46, 47, 48, 49],
                [50, 51, 52, 53, 54],
                [55, 56, 57, 58, 59],
            ],
        ],
    )


def test_delete_1D(small_array):
    np.delete(small_array, 2)
    assert np.array_equal(small_array, [1, 2, 4, 5])


def test_delete_3D(small_large_array):
    np.delete(small_large_array, 1, axis=1)
    assert np.array_equal(
        small_large_array,
        [
            [[0, 1, 2, 3, 4], [10, 11, 12, 13, 14], [15, 16, 17, 18, 19]],
            [
                [20, 21, 22, 23, 24],
                [30, 31, 32, 33, 34],
                [35, 36, 37, 38, 39],
            ],
            [
                [40, 41, 42, 43, 44],
                [50, 51, 52, 53, 54],
                [55, 56, 57, 58, 59],
            ],
        ],
    )


def test_ravel(small_large_array) -> None:
    assert np.array_equal(
        np.ravel(small_large_array, order="C"),
        np.array(
            [
                0,
                1,
                2,
                3,
                4,
                5,
                6,
                7,
                8,
                9,
                10,
                11,
                12,
                13,
                14,
                15,
                16,
                17,
                18,
                19,
                20,
                21,
                22,
                23,
                24,
                25,
                26,
                27,
                28,
                29,
                30,
                31,
                32,
                33,
                34,
                35,
                36,
                37,
                38,
                39,
                40,
                41,
                42,
                43,
                44,
                45,
                46,
                47,
                48,
                49,
                50,
                51,
                52,
                53,
                54,
                55,
                56,
                57,
                58,
                59,
            ]
        ),
    )

    assert np.array_equal(
        np.ravel(small_large_array, order="F"),
        np.array(
            [
                0,
                20,
                40,
                5,
                25,
                45,
                10,
                30,
                50,
                15,
                35,
                55,
                1,
                21,
                41,
                6,
                26,
                46,
                11,
                31,
                51,
                16,
                36,
                56,
                2,
                22,
                42,
                7,
                27,
                47,
                12,
                32,
                52,
                17,
                37,
                57,
                3,
                23,
                43,
                8,
                28,
                48,
                13,
                33,
                53,
                18,
                38,
                58,
                4,
                24,
                44,
                9,
                29,
                49,
                14,
                34,
                54,
                19,
                39,
                59,
            ]
        ),
    )


def test_take(small_large_array) -> None:
    assert np.array_equal(np.take(small_large_array, [[0, 1], [2, 3]]), np.array([[0, 1], [2, 3]]))


def test_may_not_share_memory(small_array) -> None:
    assert not np.may_share_memory(small_array, np.array([1, 2, 3]))


def test_may_share_memory(small_array) -> None:
    assert np.may_share_memory(small_array, small_array)


def test_may_share_memory_view(small_array) -> None:
    assert np.may_share_memory(small_array, small_array[1:3])



import ch5mpy as ch
from ch5mpy.h5array.chunks.repeated_array import RepeatedArray


def test_subset():
    arr = ch.arange_nd((8, 1, 6, 1))
    r_arr = RepeatedArray(arr, (8, 7, 6, 5))

    sub = r_arr[0:2, 3:7, 1:4]
    assert sub.shape == (2, 4, 3, 5)


def test_subset_larger_shape():
    arr = ch.arange_nd((4, 5))
    r_arr = RepeatedArray(arr, (3, 4, 5))

    sub = r_arr[0:1]
    assert sub.shape == (1, 4, 5)



from __future__ import annotations

import numpy as np

import numpy.typing as npt
from typing import Any
from typing import Literal
from typing import TypeVar
from typing import Iterable
from typing import TYPE_CHECKING

from ch5mpy.h5array.functions.implement import implements

if TYPE_CHECKING:
    from ch5mpy import H5Array


_T = TypeVar("_T", bound=np.generic)


@implements(np.copy)
def copy(a: H5Array[_T]) -> npt.NDArray[_T]:
    return np.array(a)


@implements(np.zeros_like)
def zeros_like(
    a: H5Array[Any],
    dtype: npt.DTypeLike | None = None,
    order: Literal["C", "F"] | None = "C",
    shape: Iterable[int] | None = None,
) -> Any:
    return np.zeros(shape=shape or a.shape, dtype=dtype or a.dtype, order=order)  # type: ignore[call-overload]


@implements(np.ones_like)
def ones_like(
    a: H5Array[Any],
    dtype: npt.DTypeLike = None,
    order: Literal["C", "F"] | None = "C",
    shape: Iterable[int] | None = None,
) -> Any:
    return np.ones(shape=shape or a.shape, dtype=dtype or a.dtype, order=order)  # type: ignore[call-overload]



from __future__ import annotations

from typing import Any, Generator, TypeVar, cast

import numpy as np
from numpy import typing as npt

from ch5mpy.indexing.selection import Selection
from ch5mpy.objects import Dataset, DatasetWrapper

_DT = TypeVar("_DT", bound=np.generic)


class IterWithFinalReordering:
    def __init__(self, gen: Generator[Any, Any, Any]):
        self.gen = gen
        self.final_reordering: tuple[npt.NDArray[np.int_] | slice, ...] | None = None

    def __iter__(self) -> Generator[Any, Any, Any]:
        self.final_reordering = yield from self.gen


def read_from_dataset(
    dataset: Dataset[_DT] | DatasetWrapper[_DT],
    selection: Selection,
    loading_array: npt.NDArray[_DT],
) -> None:
    if not dataset.size:
        if loading_array.size:
            raise ValueError("Reading from empty dataset.")
        return

    if not loading_array.size:
        return

    indexers = IterWithFinalReordering(selection.iter_indexers())
    for dataset_idx, loading_array_idx in indexers:
        # TODO : would be nice to be able to pass an array with random order in `dest_sel`
        dataset.read_direct(loading_array, source_sel=dataset_idx, dest_sel=loading_array_idx)

    if indexers.final_reordering is not None:
        loading_array[:] = loading_array[indexers.final_reordering]


def read_one_from_dataset(
    dataset: Dataset[_DT] | DatasetWrapper[_DT],
    selection: Selection,
    dtype: np.dtype[_DT],
) -> _DT:
    loading_array = np.empty((), dtype=dtype)
    read_from_dataset(dataset, selection, loading_array)
    return cast(_DT, loading_array[()])


def write_to_dataset(
    dataset: Dataset[_DT] | DatasetWrapper[_DT],
    values: npt.NDArray[_DT],
    selection: Selection,
) -> None:
    selection_shape = selection.out_shape

    if values.size == np.prod(selection_shape) and values.shape != selection_shape:
        values = values.reshape(selection_shape)

    for dataset_idx, array_idx in selection.iter_indexers(can_reorder=False):
        dataset.write_direct(values, source_sel=array_idx, dest_sel=dataset_idx)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Collection, Sequence

if TYPE_CHECKING:
    from typing_extensions import TypeGuard


def is_sequence(obj: Any) -> TypeGuard[Sequence[Any]]:
    """Is the object a sequence of objects ? (excluding strings and byte objects.)"""
    return (
        isinstance(obj, Collection)
        and hasattr(obj, "__getitem__")
        and not isinstance(obj, (str, bytes, bytearray, memoryview))
    )



from __future__ import annotations

from pathlib import Path
from typing import Any, Generator

import pytest

import ch5mpy as ch


class O_:
    def __init__(self, v: float):
        self._v = v

    def __repr__(self) -> str:
        return f"O({self._v})"

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, O_):
            return False

        return self._v == other._v


class C:
    def __init__(self, l_: list[int]):
        self.l = l_

    def __repr__(self) -> str:
        return f"C({self.l})"

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, C):
            return False

        return self.l == other.l

    def __h5_write__(self, values: ch.H5Dict) -> None:
        values["l"] = self.l

    @classmethod
    def __h5_read__(cls, values: ch.H5Dict) -> C:
        return C(list(values["l"]))


@pytest.fixture
def h5_list() -> Generator[ch.H5List, None, None]:
    data = [1.0, 2, C([1, 2, 3]), "4.", O_(5.0)]

    with ch.File("h5_list.h5", ch.H5Mode.WRITE_TRUNCATE) as h5_file:
        ch.H5List.write(data, h5_file, "data")

    yield ch.H5List.read("h5_list.h5", "data", mode=ch.H5Mode.READ_WRITE)

    Path("h5_list.h5").unlink()


def test_list_repr(h5_list):
    assert repr(h5_list) == "H5List[1.0, 2, C([1, 2, 3]), '4.', O(5.0)]"


def test_list_should_read_custom_object(h5_list):
    assert isinstance(h5_list[4], O_)


def test_list_should_read_custom_object_with_method(h5_list):
    assert isinstance(h5_list[2], C)


def test_list_should_convert_to_regular_list(h5_list):
    lst = h5_list.copy()
    assert lst == [1.0, 2, C([1, 2, 3]), "4.", O_(5.0)]


def test_list_get_negative_index(h5_list):
    assert h5_list[-2] == "4."


def test_can_append_value_to_list(h5_list):
    h5_list.append(-1)
    assert h5_list[5] == -1



import pickle
import numpy as np

from ch5mpy.write import write_object


def test_should_pickle_dataset(group):
    write_object(group, "something", [1, 2, 3])

    pickled_obj = pickle.dumps(group["something"], protocol=pickle.HIGHEST_PROTOCOL)
    unpickled_obj = pickle.loads(pickled_obj)

    assert np.array_equal(unpickled_obj[:], [1, 2, 3])


def test_should_pickle_dict(group):
    write_object(group, "something", {"a": 1, "b": "2"})

    pickled_obj = pickle.dumps(group["something"], protocol=pickle.HIGHEST_PROTOCOL)
    unpickled_obj = pickle.loads(pickled_obj)

    assert unpickled_obj["a"][()] == 1
    assert unpickled_obj["b"][()] == b"2"



from __future__ import annotations

from pathlib import Path
from typing import Any, Generic, Iterator, TypeVar, cast

from ch5mpy.dict import H5Dict
from ch5mpy.names import H5Mode
from ch5mpy.objects import File, Group, H5Object
from ch5mpy.read import read_object
from ch5mpy.write import write_object

_T = TypeVar("_T")


def _get_group(file: File | Group, name: str) -> tuple[File | Group, str]:
    lst = name.rsplit("/", 1)

    if len(lst) == 1:
        return file, lst[0]

    return file.require_group(lst[0]), lst[1]


class H5List(H5Object, Generic[_T]):
    """Class for managing lists backed on h5 files."""

    # region magic methods
    def __init__(self, file: File | Group):
        super().__init__(file)
        self._iter_index = 0

    def __repr__(self) -> str:
        if self.is_closed:
            return "Closed H5List[...]"

        if len(self) > 6:
            return f"H5List[{self[0]}, {self[1]}, {self[2]}, ..., " f"{self[-3]}, {self[-2]}, {self[-1]}]"

        return f"H5List[{', '.join([repr(e) for e in self])}]"

    def __len__(self) -> int:
        return len(self._file.keys())

    def __getitem__(self, item: int) -> _T:
        if item < 0:
            item = len(self) + item

        return cast(_T, read_object(self._file[str(item)]))

    def __iter__(self) -> Iterator[_T]:
        self._iter_index = 0
        return self

    def __next__(self) -> _T:
        if self._iter_index >= len(self):
            raise StopIteration

        obj = cast(_T, read_object(self._file[str(self._iter_index)]))
        self._iter_index += 1

        return obj

    # endregion

    # region class methods
    @classmethod
    def read(cls, path: str | Path | File | Group, name: str = "", mode: H5Mode = H5Mode.READ) -> H5List[Any]:
        if isinstance(path, (str, Path)):
            path = File(path, mode=mode)

        src = path[name]

        if not isinstance(src, Group):
            raise ValueError(f"Cannot read H5List from '{type(src)}' object, " f"should be a 'Group'.")

        src_type = src.attrs.get("__h5_type__", "")

        if src_type != "list":
            raise ValueError(f"Cannot read H5List from Group marked as type " f"'{src_type}', should be 'list'.")

        return H5List(src)

    @classmethod
    def write(cls, lst: list[Any], path: str | Path | File | Group, name: str = "") -> None:
        """
        Write a list to a .h5 file as a H5List.

        Args:
            lst: a list to write.
            path: Either a path to the .h5 file or an open ch5mpy.File or
                ch5mpy.Group object.
            name: The group member name inside the .h5 file to use.
        """
        if isinstance(path, (str, Path)):
            path = File(path, mode=H5Mode.READ_WRITE)

        if path.file.mode == H5Mode.READ:
            raise ValueError("Cannot write to h5 file open in 'r' mode.")

        dest, group_name = _get_group(path, name)

        if group_name in dest.keys():
            raise ValueError(f"An object with name '{group_name}' already exists.")

        dest = dest.create_group(group_name)
        dest.attrs["__h5_type__"] = "list"

        for index, obj in enumerate(lst):
            write_object(dest, str(index), obj, chunks=True)

    # endregion

    # region methods
    def copy(self) -> Any:
        return [e for e in self]

    def to_dict(self) -> H5Dict[_T]:
        return H5Dict(self._file)

    def append(self, value: None) -> None:
        write_object(self._file, str(len(self)), value, chunks=True)

    # endregion



from __future__ import annotations

from functools import partial
from pathlib import Path
from typing import Any, Callable, Protocol, cast

import numpy as np
import numpy.typing as npt

import ch5mpy
from ch5mpy.write import _store_dataset


_NoValue = object()


class ArrayCreationFunc:
    # region magic methods
    def __init__(self, name: str) -> None:
        self._fill_value: Any = _NoValue
        self._name = name

    def __repr__(self) -> str:
        return f"<function ch5mpy.{self._name} at {hex(id(self))}>"

    def __call__(
        self,
        shape: int | tuple[int, ...],
        name: str,
        loc: str | Path | ch5mpy.File | ch5mpy.Group,
        dtype: npt.DTypeLike = np.float64,
        chunks: bool | tuple[int, ...] = True,
        maxshape: int | tuple[int | None, ...] | None = None,
    ) -> ch5mpy.H5Array[Any]:
        if self._fill_value == _NoValue:
            raise RuntimeError

        shape = shape if isinstance(shape, tuple) else (shape,)

        if not isinstance(loc, (ch5mpy.File, ch5mpy.Group)):
            loc = ch5mpy.File(loc, mode=ch5mpy.H5Mode.READ_WRITE_CREATE)

        dset = _store_dataset(
            loc, name, shape=shape, dtype=dtype, chunks=chunks, maxshape=maxshape, fill_value=self._fill_value
        )

        return ch5mpy.H5Array(dset)

    # endregion

    # region methods
    def anonymous(
        self,
        shape: int | tuple[int, ...],
        dtype: npt.DTypeLike = np.float64,
        chunks: bool | tuple[int, ...] = True,
        maxshape: int | tuple[int | None, ...] | None = None,
    ) -> partial[ch5mpy.H5Array[Any]]:
        return partial(self.__call__, shape=shape, dtype=dtype, chunks=chunks, maxshape=maxshape)

    # endregion


class ArrayCreationFuncWithFill(ArrayCreationFunc):
    # region magic methods
    def __init__(self, name: str, fill_value: Any):
        super().__init__(name)
        self._fill_value = fill_value

    def __call__(
        self,
        shape: int | tuple[int, ...],
        name: str,
        loc: str | Path | ch5mpy.File | ch5mpy.Group,
        dtype: npt.DTypeLike = np.float64,
        chunks: bool | tuple[int, ...] = True,
        maxshape: int | tuple[int | None, ...] | None = None,
    ) -> ch5mpy.H5Array[Any]:
        return super().__call__(shape, name, loc, dtype, chunks, maxshape)

    # endregion


class _ArrayCreationFuncProtocol(Protocol):
    acf: ArrayCreationFunc

    def __call__(
        self,
        shape: int | tuple[int, ...],
        name: str,
        loc: str | Path | ch5mpy.File | ch5mpy.Group,
        dtype: npt.DTypeLike = np.float64,
        chunks: bool | tuple[int, ...] = True,
        maxshape: int | tuple[int | None, ...] | None = None,
    ) -> ch5mpy.H5Array[Any]:
        ...

    def anonymous(
        self,
        shape: int | tuple[int, ...],
        dtype: npt.DTypeLike = np.float64,
        chunks: bool | tuple[int, ...] = True,
        maxshape: int | tuple[int | None, ...] | None = None,
    ) -> partial[ch5mpy.H5Array[Any]]:
        ...


def with_anonymous(acf: ArrayCreationFunc) -> Callable[[Callable[..., Any]], _ArrayCreationFuncProtocol]:
    """Add a `.anonymous()` method to a function for delaying the array creation to when the location is known."""

    def decorator(func: Callable[..., Any]) -> _ArrayCreationFuncProtocol:
        setattr(func, "acf", acf)
        setattr(func, "anonymous", acf.anonymous)
        return cast(_ArrayCreationFuncProtocol, func)

    return decorator


@with_anonymous(ArrayCreationFuncWithFill("empty", None))
def empty(
    shape: int | tuple[int, ...],
    name: str,
    loc: str | Path | ch5mpy.File | ch5mpy.Group,
    dtype: npt.DTypeLike = np.float64,
    chunks: bool | tuple[int, ...] = True,
    maxshape: int | tuple[int | None, ...] | None = None,
) -> ch5mpy.H5Array[Any]:
    return empty.acf(shape, name, loc, dtype, chunks, maxshape)


@with_anonymous(ArrayCreationFuncWithFill("zeros", 0))
def zeros(
    shape: int | tuple[int, ...],
    name: str,
    loc: str | Path | ch5mpy.File | ch5mpy.Group,
    dtype: npt.DTypeLike = np.float64,
    chunks: bool | tuple[int, ...] = True,
    maxshape: int | tuple[int | None, ...] | None = None,
) -> ch5mpy.H5Array[Any]:
    return zeros.acf(shape, name, loc, dtype, chunks, maxshape)


@with_anonymous(ArrayCreationFuncWithFill("ones", 1))
def ones(
    shape: int | tuple[int, ...],
    name: str,
    loc: str | Path | ch5mpy.File | ch5mpy.Group,
    dtype: npt.DTypeLike = np.float64,
    chunks: bool | tuple[int, ...] = True,
    maxshape: int | tuple[int | None, ...] | None = None,
) -> ch5mpy.H5Array[Any]:
    return ones.acf(shape, name, loc, dtype, chunks, maxshape)


@with_anonymous(ArrayCreationFunc("full"))
def full(
    shape: int | tuple[int, ...],
    fill_value: Any,
    name: str,
    loc: str | Path | ch5mpy.File | ch5mpy.Group,
    dtype: npt.DTypeLike = np.float64,
    chunks: bool | tuple[int, ...] = True,
    maxshape: int | tuple[int | None, ...] | None = None,
) -> ch5mpy.H5Array[Any]:
    full.acf._fill_value = fill_value
    return full.acf(shape, name, loc, dtype, chunks, maxshape)



try:
    from importlib import metadata
except ImportError:  # for Python<3.8
    import importlib_metadata as metadata  # type: ignore[no-redef]

from ch5mpy.attributes import AttributeManager
from ch5mpy.dict import H5Dict
from ch5mpy.h5array.array import H5Array
from ch5mpy.h5array.creation_routines import empty, full, ones, zeros
from ch5mpy.list import H5List
from ch5mpy.names import H5Mode
from ch5mpy.np import arange_nd
from ch5mpy.objects import Dataset, File, Group
from ch5mpy.options import options, set_options
from ch5mpy.read import read_object
from ch5mpy.write import (
    write_dataset,
    write_datasets,
    write_object,
    write_objects,
)

__all__ = [
    "File",
    "Group",
    "Dataset",
    "H5Dict",
    "H5List",
    "H5Array",
    "AttributeManager",
    "write_dataset",
    "write_datasets",
    "write_object",
    "write_objects",
    "read_object",
    "H5Mode",
    "arange_nd",
    "empty",
    "zeros",
    "ones",
    "full",
    "options",
    "set_options",
]

__version__ = metadata.version("ch5mpy")



from ch5mpy.indexing.slice import FullSlice


def test_true_stop_step1():
    f = FullSlice(0, 3, 1, 10)
    assert f.true_stop == 2


def test_true_stop_step3():
    f = FullSlice(5, 13, 3, 20)
    assert f.true_stop == 11


def test_length():
    f = FullSlice(0, 3, 1, 10)
    assert len(f) == 3



from pathlib import Path

import numpy as np
import pytest

import ch5mpy


def test_should_get_shape(array):
    assert array.shape == (10, 10)


def test_should_get_dtype(array):
    assert array.dtype == np.float64


def test_should_print_repr(array):
    assert (
        repr(array) == "H5Array([[0.0, 1.0, 2.0, ..., 7.0, 8.0, 9.0],\n"
        "         [10.0, 11.0, 12.0, ..., 17.0, 18.0, 19.0],\n"
        "         [20.0, 21.0, 22.0, ..., 27.0, 28.0, 29.0],\n"
        "         ...,\n"
        "         [70.0, 71.0, 72.0, ..., 77.0, 78.0, 79.0],\n"
        "         [80.0, 81.0, 82.0, ..., 87.0, 88.0, 89.0],\n"
        "         [90.0, 91.0, 92.0, ..., 97.0, 98.0, 99.0]], shape=(10, 10), dtype=float64)"
    )


def test_should_convert_to_numpy_array(array):
    assert type(np.asarray(array)) == np.ndarray
    assert np.array_equal(np.asarray(array), np.arange(100).reshape((10, 10)))


def test_should_pass_numpy_ufunc(array):
    arr_2 = np.multiply(array, 2)
    assert np.array_equal(arr_2, np.arange(100).reshape((10, 10)) * 2)


def test_should_work_with_magic_operations(array):
    arr_2 = array + 2
    assert np.array_equal(arr_2, np.arange(100).reshape((10, 10)) + 2)


def test_should_sum_all_values(array):
    s = np.sum(array)
    assert s == 4950


def test_should_sum_along_axis(array):
    s = np.sum(array, axis=0)
    assert np.array_equal(s, np.array([450, 460, 470, 480, 490, 500, 510, 520, 530, 540]))


def test_should_add_inplace(array):
    array += 1
    assert np.array_equal(array, np.arange(100).reshape((10, 10)) + 1)


def test_should_add_to_view(array):
    view = array[:, 0]
    res = view + 1

    assert type(res) == np.ndarray
    assert np.array_equal(res, np.array([1, 11, 21, 31, 41, 51, 61, 71, 81, 91]))


def test_should_add_inplace_to_view(array):
    v = array[1:3, 5:8]
    v += 1

    assert np.array_equal(
        array,
        np.array(
            [
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [10, 11, 12, 13, 14, 16, 17, 18, 18, 19],
                [20, 21, 22, 23, 24, 26, 27, 28, 28, 29],
                [30, 31, 32, 33, 34, 35, 36, 37, 38, 39],
                [40, 41, 42, 43, 44, 45, 46, 47, 48, 49],
                [50, 51, 52, 53, 54, 55, 56, 57, 58, 59],
                [60, 61, 62, 63, 64, 65, 66, 67, 68, 69],
                [70, 71, 72, 73, 74, 75, 76, 77, 78, 79],
                [80, 81, 82, 83, 84, 85, 86, 87, 88, 89],
                [90, 91, 92, 93, 94, 95, 96, 97, 98, 99],
            ]
        ),
    )


def test_should_add_inplace_to_view_2(array):
    array[[0, 2, 3], 5:7] += 1

    assert np.array_equal(
        array,
        np.array(
            [
                [0, 1, 2, 3, 4, 6, 7, 7, 8, 9],
                [10, 11, 12, 13, 14, 15, 16, 17, 18, 19],
                [20, 21, 22, 23, 24, 26, 27, 27, 28, 29],
                [30, 31, 32, 33, 34, 36, 37, 37, 38, 39],
                [40, 41, 42, 43, 44, 45, 46, 47, 48, 49],
                [50, 51, 52, 53, 54, 55, 56, 57, 58, 59],
                [60, 61, 62, 63, 64, 65, 66, 67, 68, 69],
                [70, 71, 72, 73, 74, 75, 76, 77, 78, 79],
                [80, 81, 82, 83, 84, 85, 86, 87, 88, 89],
                [90, 91, 92, 93, 94, 95, 96, 97, 98, 99],
            ]
        ),
    )


# def test_large_array(large_array):
#     large_array += 2


def test_should_get_single_element(array):
    assert array[1, 2] == 12


def test_should_get_whole_dset(array):
    assert np.array_equal(array[:], array)
    assert np.array_equal(array[()], array)


def test_should_print_view_repr(array):
    sub_arr = array[2:4, [0, 2, 3]]
    assert str(sub_arr) == "[[20.0 22.0 23.0],\n" " [30.0 32.0 33.0]]\n"


def test_should_get_view(array):
    sub_arr = array[2:4, [0, 2, 3]]
    assert np.array_equal(sub_arr, np.array([[20, 22, 23], [30, 32, 33]]))


def test_should_get_view_from_view(array):
    sub_arr = array[2:4, [0, 2, 3]]
    sub_sub_arr = sub_arr[1, [1, 2]]
    assert np.array_equal(sub_sub_arr, np.array([32, 33]))


def test_should_get_single_value_from_view(array):
    assert array[2:4, [0, 2, 3]][0, 1] == 22


def test_should_subset_from_boolean_array(array):
    subset = array[np.array([True, False, True, False, False, False, False, False, False, False])]
    assert np.array_equal(subset, np.array([[0, 1, 2, 3, 4, 5, 6, 7, 8, 9], [20, 21, 22, 23, 24, 25, 26, 27, 28, 29]]))


def test_should_subset_from_2_boolean_arrays(array):
    subset = array[
        [True, False, True, False, False, False, False, False, False, False],
        [True, False, True, False, False, False, False, False, False, False],
    ]
    assert np.array_equal(subset, np.array([0, 22]))


def test_should_set_value_in_array(array):
    array[5, 7] = -1
    assert array[5, 7] == -1


def test_should_set_values_in_array_in_correct_order(array):
    array[:, [2, 3]][[4, 0, 2]] = np.array([[-1, -2], [-3, -4], [-5, -6]])
    assert np.array_equal(
        array,
        np.array(
            [
                [0, 1, -3, -4, 4, 5, 6, 7, 8, 9],
                [10, 11, 12, 13, 14, 15, 16, 17, 18, 19],
                [20, 21, -5, -6, 24, 25, 26, 27, 28, 29],
                [30, 31, 32, 33, 34, 35, 36, 37, 38, 39],
                [40, 41, -1, -2, 44, 45, 46, 47, 48, 49],
                [50, 51, 52, 53, 54, 55, 56, 57, 58, 59],
                [60, 61, 62, 63, 64, 65, 66, 67, 68, 69],
                [70, 71, 72, 73, 74, 75, 76, 77, 78, 79],
                [80, 81, 82, 83, 84, 85, 86, 87, 88, 89],
                [90, 91, 92, 93, 94, 95, 96, 97, 98, 99],
            ]
        ),
    )


def test_should_set_value_in_view(array):
    sub_arr = array[2:4, [0, 2, 3]]
    sub_arr[1, [1, 2]] = [-2, -3]
    assert np.array_equal(array[3, [2, 3]], [-2, -3])


def test_apply_all_function(array):
    assert not np.all(array)

    array += 1
    assert np.all(array)


def test_apply_any_function(array):
    assert np.any(array)


def test_should_find_value_in_array(small_large_array):
    with ch5mpy.options(max_memory=str(3 * small_large_array.dtype.itemsize)):
        assert 10 in small_large_array


def test_should_not_find_missing_value_in_array(small_large_array):
    with ch5mpy.options(max_memory=str(3 * small_large_array.dtype.itemsize)):
        assert 1000 not in small_large_array


def test_should_find_value_in_view(small_array) -> None:
    view = small_array[2:]
    assert 4 in view


def test_should_not_find_missing_value_in_view(small_array) -> None:
    view = small_array[2:]
    assert 1 not in view


def test_subset_0d(array):
    subset = array[[]]
    assert subset.shape == (0, 10)
    assert repr(subset) == "H5Array([], shape=(0, 10), dtype=float64)"


def test_subset_1d_0d(array):
    subset = array[np.ix_([0, 1, 2], [])]
    assert subset.shape == (3, 0)
    assert subset.size == 0
    assert repr(subset) == "H5Array([], shape=(3, 0), dtype=float64)"


def test_subset_1d(array):
    subset = array[0]
    assert subset.ndim == 1


def test_subset_1d_column(array):
    subset = array[:, 0]
    assert subset.shape == (10,)


def test_subset_2d(array):
    subset = array[[0, 2], [[0]]]
    assert subset.ndim == 2


def test_view_should_convert_to_numpy(array):
    subset = array[[0, 2], [[0]]]
    subset = subset.copy()
    assert np.array_equal(subset, [[0.0, 20.0]])


def test_view_should_convert_to_numpy_2(small_large_array):
    subset = small_large_array[[[0, 1], [1, 0]], [[3, 2], [1, 0]], [[0, 1], [2, 3]]]
    subset = subset.copy()
    assert np.array_equal(subset, [[15, 31], [27, 3]])


def test_view_as_object_dtype_should_convert_to_numpy(str_array):
    subset = str_array[0:1].astype(object)
    subset = subset.copy()
    assert np.array_equal(subset, [b"a"])


@pytest.mark.parametrize("subset, shape", [[(slice(None), 0), (0,)], [(None, slice(None), 0, None), (1, 0, 1)]])
def test_subset_from_empty_array(empty_array, subset, shape):
    subset = empty_array[subset]
    assert subset.shape == shape


def test_empty_array_should_convert_to_numpy_array(empty_array):
    assert np.array_equal(np.array(empty_array), np.empty((0, 1)))


@pytest.mark.parametrize("subset, array", [[(slice(None), 0), np.empty(0)], [[], np.empty((0, 1))]])
def test_empty_array_subset_should_convert_to_numpy_array(empty_array, subset, array):
    subset = empty_array[subset]
    assert np.array_equal(np.array(subset), array)


def test_array_subset_ix(array):
    assert array[np.ix_([5], [5])] == 55


def test_array_subset_2d(array):
    subarr = array[[0]]
    assert subarr.ndim == 2
    assert repr(subarr) == "H5Array([[0.0, 1.0, 2.0, ..., 7.0, 8.0, 9.0]], shape=(1, 10), dtype=float64)"
    assert subarr[0].ndim == 1


def test_array_subset_3d(array):
    subarr = array[[[0]]]
    assert subarr.ndim == 3
    assert subarr[0].ndim == 2
    assert subarr[0, 0].ndim == 1
    assert repr(subarr) == "H5Array([[[0.0, 1.0, 2.0, ..., 7.0, 8.0, 9.0]]], shape=(1, 1, 10), dtype=float64)"


def test_array_should_get_one_element(array):
    subarr = array[0, 0]
    assert isinstance(subarr, np.float_)


def test_array_should_get_array_of_one_element_in_1d(array):
    subarr = array[[0], [0]]
    assert isinstance(subarr, ch5mpy.H5Array)
    assert subarr.shape == (1,)


def test_array_should_get_array_of_one_element_in_2d(array):
    subarr = array[[0], [[0]]]
    assert isinstance(subarr, ch5mpy.H5Array)
    assert subarr.shape == (1, 1)


def test_array_should_get_array_of_multiple_elements_in_2d(array):
    subarr = array[[[0], [2], [5]], [[0]]]
    assert isinstance(subarr, ch5mpy.H5Array)
    assert subarr.shape == (3, 1)
    assert subarr[0].shape == (1,)
    assert repr(subarr) == "H5Array([[0.0],\n" "         [20.0],\n" "         [50.0]], shape=(3, 1), dtype=float64)"


def test_array_should_get_array_in_1d_from_slice(array):
    subarr = array[:3]
    subsubarr = subarr[0]
    assert isinstance(subsubarr, ch5mpy.H5Array)
    assert subsubarr.shape == (10,)


def test_array_type_casting(array):
    assert array.astype(int).dtype == np.int64


def test_view_type_casting(array):
    subarr = array[:5, :5]
    assert subarr.astype(int).shape == (5, 5)


def test_subset_newaxis(small_array):
    new_array = small_array[:, None]
    assert new_array.shape == (5, 1)


def test_subset_newaxis_shape(small_array):
    new_array = small_array[:, None]

    subarr = new_array[0]
    assert subarr.shape == (1,)

    subsubarr = subarr[0]
    assert isinstance(subsubarr, np.float_)


def test_subset_multiple_newaxis_repr(small_array):
    new_array = small_array[None, None]

    assert new_array.shape == (1, 1, 5)
    assert repr(new_array) == "H5Array([[[1.0, 2.0, 3.0, 4.0, 5.0]]], shape=(1, 1, 5), dtype=float64)"


def test_newaxis_to_numpy_array(small_array):
    new_array = small_array[:, None]
    assert np.array(new_array).shape == (5, 1)


def test_setitem(array):
    array[0, [0, 1]] = np.array([[-1, -2]])

    ref = np.arange(100.0).reshape((10, 10))
    ref[0, [0, 1]] = [-1, -2]

    assert np.array_equal(array, ref)


def test_setitem_row(array):
    array[[[1]], [[0, 1]]] = np.array([[-1, -2]])


def test_setitem_column(array):
    v = -1 * np.ones(5)[:, None]
    array[[[0], [1], [2], [3], [4]], [[0]]] = v


def test_get_in_random_order(array):
    v = array[np.ix_([5, 1, 2], [9, 8, 7])]
    a = np.array(v)
    assert np.array_equal(a, [[59, 58, 57], [19, 18, 17], [29, 28, 27]])


def test_get_in_random_order_1d(small_array):
    v = small_array[[1, 3, 4, 0, 2]]
    a = np.array(v)
    assert np.array_equal(a, [2, 4, 5, 1, 3])


def test_get_in_random_order_2d(array):
    v = array[np.ix_([1, 4, 1, 3, 1, 2], [2, 1, 0])]
    a = np.array(v)
    assert np.array_equal(
        a,
        np.array(
            [
                [12.0, 11.0, 10.0],
                [42.0, 41.0, 40.0],
                [12.0, 11.0, 10.0],
                [32.0, 31.0, 30.0],
                [12.0, 11.0, 10.0],
                [22.0, 21.0, 20.0],
            ]
        ),
    )


def test_inplace_type_casting(array: ch5mpy.H5Array):
    array.astype(str, inplace=True)
    assert array.dtype == np.dtype("<U4")


# FIXME
@pytest.mark.xfail
def test_repr_3d_array():
    path = Path("f.h5")
    if path.exists():
        path.unlink()

    array = ch5mpy.ones((10, 10, 10), "array", "f.h5")
    assert (
        repr(array) == "H5Array([[[1.0, 1.0, 1.0, ..., 1.0, 1.0, 1.0],\n"
        "          [1.0, 1.0, 1.0, ..., 1.0, 1.0, 1.0],\n"
        "          [1.0, 1.0, 1.0, ..., 1.0, 1.0, 1.0],\n"
        "          ...,\n"
        "          [1.0, 1.0, 1.0, ..., 1.0, 1.0, 1.0],\n"
        "          [1.0, 1.0, 1.0, ..., 1.0, 1.0, 1.0],\n"
        "          [1.0, 1.0, 1.0, ..., 1.0, 1.0, 1.0]],\n"
        "\n"
        "         [[1.0, 1.0, 1.0, ..., 1.0, 1.0, 1.0],\n"
        "          [1.0, 1.0, 1.0, ..., 1.0, 1.0, 1.0],\n"
        "          [1.0, 1.0, 1.0, ..., 1.0, 1.0, 1.0],\n"
        "          ...,\n"
        "          [1.0, 1.0, 1.0, ..., 1.0, 1.0, 1.0],\n"
        "          [1.0, 1.0, 1.0, ..., 1.0, 1.0, 1.0],\n"
        "          [1.0, 1.0, 1.0, ..., 1.0, 1.0, 1.0]],\n"
        "\n"
        "         [[1.0, 1.0, 1.0, ..., 1.0, 1.0, 1.0],\n"
        "          [1.0, 1.0, 1.0, ..., 1.0, 1.0, 1.0],\n"
        "          [1.0, 1.0, 1.0, ..., 1.0, 1.0, 1.0],\n"
        "          ...,\n"
        "          [1.0, 1.0, 1.0, ..., 1.0, 1.0, 1.0],\n"
        "          [1.0, 1.0, 1.0, ..., 1.0, 1.0, 1.0],\n"
        "          [1.0, 1.0, 1.0, ..., 1.0, 1.0, 1.0]],\n"
        "\n"
        "         ...,\n"
        "\n"
        "         [[1.0, 1.0, 1.0, ..., 1.0, 1.0, 1.0],\n"
        "          [1.0, 1.0, 1.0, ..., 1.0, 1.0, 1.0],\n"
        "          [1.0, 1.0, 1.0, ..., 1.0, 1.0, 1.0],\n"
        "          ...,\n"
        "          [1.0, 1.0, 1.0, ..., 1.0, 1.0, 1.0],\n"
        "          [1.0, 1.0, 1.0, ..., 1.0, 1.0, 1.0],\n"
        "          [1.0, 1.0, 1.0, ..., 1.0, 1.0, 1.0]],\n"
        "\n"
        "         [[1.0, 1.0, 1.0, ..., 1.0, 1.0, 1.0],\n"
        "          [1.0, 1.0, 1.0, ..., 1.0, 1.0, 1.0],\n"
        "          [1.0, 1.0, 1.0, ..., 1.0, 1.0, 1.0],\n"
        "          ...,\n"
        "          [1.0, 1.0, 1.0, ..., 1.0, 1.0, 1.0],\n"
        "          [1.0, 1.0, 1.0, ..., 1.0, 1.0, 1.0],\n"
        "          [1.0, 1.0, 1.0, ..., 1.0, 1.0, 1.0]],\n"
        "\n"
        "         [[1.0, 1.0, 1.0, ..., 1.0, 1.0, 1.0],\n"
        "          [1.0, 1.0, 1.0, ..., 1.0, 1.0, 1.0],\n"
        "          [1.0, 1.0, 1.0, ..., 1.0, 1.0, 1.0],\n"
        "          ...,\n"
        "          [1.0, 1.0, 1.0, ..., 1.0, 1.0, 1.0],\n"
        "          [1.0, 1.0, 1.0, ..., 1.0, 1.0, 1.0],\n"
        "          [1.0, 1.0, 1.0, ..., 1.0, 1.0, 1.0]]], shape=(10, 10, 10), dtype=float64)"
    )
    path.unlink()


def test_can_flatten(small_large_array: ch5mpy.H5Array) -> None:
    assert small_large_array.flatten().shape == (60,)


def test_can_hash(small_array: ch5mpy.H5Array) -> None:
    hash1 = hash(small_array)
    small_array[0] = -1
    hash2 = hash(small_array)
    assert hash1 != hash2



from __future__ import annotations

from typing import Any, TypeVar

import numpy as np
import numpy.typing as npt

import ch5mpy
import ch5mpy.indexing as ci
from ch5mpy import Dataset
from ch5mpy._typing import NP_FUNC, SELECTOR
from ch5mpy.h5array.array import as_array
from ch5mpy.h5array.io import read_from_dataset, read_one_from_dataset, write_to_dataset
from ch5mpy.objects import DatasetWrapper


_T = TypeVar("_T", bound=np.generic)


class H5ArrayView(ch5mpy.H5Array[_T]):
    """A view on a H5Array."""

    # region magic methods
    def __init__(self, dset: Dataset[_T] | DatasetWrapper[_T], sel: ci.Selection):
        super().__init__(dset)
        self._selection = sel

    def __getitem__(self, index: SELECTOR | tuple[SELECTOR, ...]) -> _T | ch5mpy.H5Array[_T]:
        selection = ci.Selection.from_selector(index, self.shape)

        if selection.is_empty:
            return H5ArrayView(dset=self._dset, sel=self._selection)

        selection = selection.cast_on(self._selection)

        if selection.is_empty:
            return ch5mpy.H5Array(self._dset)

        if selection.out_shape == ():
            return read_one_from_dataset(self._dset, selection, self.dtype)

        return H5ArrayView(dset=self._dset, sel=selection)

    def __setitem__(self, index: SELECTOR | tuple[SELECTOR, ...], value: Any) -> None:
        selection = ci.Selection.from_selector(index, self.shape)
        write_to_dataset(self._dset, as_array(value, self.dtype), selection.cast_on(self._selection))

    def __len__(self) -> int:
        return self.shape[0]

    def _inplace(self, func: NP_FUNC, value: Any) -> H5ArrayView[_T]:
        if np.issubdtype(self.dtype, str):
            raise TypeError("Cannot perform inplace operation on str H5Array.")

        # special case : 0D array
        if self.shape == ():
            self._dset[:] = func(self._dset[:], value)
            return self

        # general case : 1D+ array
        for index, chunk in self.iter_chunks():
            func(chunk, value, out=chunk)

            for dest_sel, source_sel in ci.Selection(index, self.shape).cast_on(self._selection).iter_indexers():
                self._dset.write_direct(chunk, source_sel=source_sel, dest_sel=dest_sel)

        return self

    # endregion

    # region interface
    def __array__(self, dtype: npt.DTypeLike | None = None) -> npt.NDArray[Any]:
        loading_array = np.empty(
            self._selection.out_shape,
            dtype or self.dtype,
        )
        read_from_dataset(self._dset, self._selection, loading_array)

        return loading_array.reshape(self.shape)

    # endregion

    # region attributes
    @property
    def shape(self) -> tuple[int, ...]:
        return self._selection.out_shape

    # endregion

    # region methods
    def astype(self, dtype: npt.DTypeLike, inplace: bool = False) -> H5ArrayView[Any]:
        """
        Cast an H5Array to a specified dtype.
        This does not perform a copy, it returns a wrapper around the underlying H5 dataset.
        """
        if inplace:
            raise TypeError("Cannot cast inplace a view of an H5Array.")

        if np.issubdtype(dtype, str) and (np.issubdtype(self._dset.dtype, str) or self._dset.dtype == object):
            return H5ArrayView(self._dset.asstr(), sel=self._selection)

        return H5ArrayView(self._dset.astype(dtype), sel=self._selection)

    def maptype(self, otype: type[Any]) -> H5ArrayView[Any]:
        """
        Cast an H5Array to any object type.
        This extends H5Array.astype() to any type <T>, where it is required that an object <T> can be constructed as
        T(v) for any value <v> in the dataset.
        """
        return H5ArrayView(self._dset.maptype(otype), sel=self._selection)

    def read_direct(
        self,
        dest: npt.NDArray[_T],
        source_sel: tuple[int | slice, ...],
        dest_sel: tuple[int | slice, ...],
    ) -> None:
        dataset = self._dset.asstr() if np.issubdtype(self.dtype, str) else self._dset
        read_from_dataset(
            dataset,
            ci.Selection(
                (ci.as_indexer(sl, max=axis_shape) for sl, axis_shape in zip(source_sel, self.shape)),
                shape=self.shape,
            ).cast_on(self._selection),
            dest[dest_sel],
        )

    # endregion



from __future__ import annotations

from enum import Enum, auto
from functools import partial
from numbers import Number
from typing import TYPE_CHECKING, Any, Iterable, Union

import numpy as np
import numpy.typing as npt
from numpy import _NoValue as NoValue  # type: ignore[attr-defined]

import ch5mpy
from ch5mpy._typing import NP_FUNC
from ch5mpy.h5array.chunks.iter import iter_chunks_2
from ch5mpy.indexing import FullSlice, SingleIndex, map_slice

if TYPE_CHECKING:
    from ch5mpy import H5Array


WHERE_SELECTION = Union[npt.NDArray[np.bool_], tuple[()]]


class ApplyOperation(Enum):
    set = auto()
    iadd = auto()
    imul = auto()
    iand = auto()
    ior = auto()


class MaskWhere:
    def __init__(
        self,
        where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue,
        shape: tuple[int, ...],
    ):
        # TODO : make memory efficient (avoid broadcast and compute on the fly)
        self._where = None if where in (True, NoValue) else np.broadcast_to(where, shape)  # type: ignore[arg-type]

    def __repr__(self) -> str:
        return f"MaskWhere({self._where})"

    def __getitem__(self, item: tuple[Any, ...] | slice) -> WHERE_SELECTION:
        if self._where is None:
            return ()

        return self._where[item]


def _get_output_array(
    out: H5Array[Any] | npt.NDArray[Any] | None,
    shape: tuple[int, ...],
    axis: tuple[int, ...],
    keepdims: bool,
    dtype: npt.DTypeLike | None,
    initial: int | float | complex | NoValue,
) -> H5Array[Any] | npt.NDArray[Any]:
    if keepdims:
        expected_shape = tuple(s if i not in axis else 1 for i, s in enumerate(shape))

    else:
        expected_shape = tuple(s for i, s in enumerate(shape) if i not in axis)

    if out is not None:
        ndim = len(expected_shape)
        if out.ndim != ndim:
            raise ValueError(f"Output array has the wrong number of dimensions: Found {out.ndim} but expected {ndim}")

        if out.shape != expected_shape:
            raise ValueError(f"Output array has the wrong shape: Found {out.shape} but expected {expected_shape}")

        if initial is not NoValue:
            out[()] = initial

        return out

    if initial is NoValue:
        return np.empty(shape=expected_shape, dtype=dtype)

    return np.full(shape=expected_shape, fill_value=initial, dtype=dtype)


def _get_output_array_2(
    out: H5Array[Any] | npt.NDArray[Any] | None,
    a_shape: tuple[int, ...],
    b_shape: tuple[int, ...],
    dtype: npt.DTypeLike | None,
    default: Any,
) -> H5Array[Any] | npt.NDArray[Any]:
    expected_shape = np.broadcast_shapes(a_shape, b_shape)

    if out is not None:
        ndim = len(expected_shape)
        if out.ndim != ndim:
            raise ValueError(f"Output array has the wrong number of dimensions: Found {out.ndim} but expected {ndim}")

        if out.shape != expected_shape:
            raise ValueError(f"Output array has the wrong shape: Found {out.shape} but expected {expected_shape}")

        return out

    if default is None:
        return np.empty(shape=expected_shape, dtype=dtype)

    return np.full(shape=expected_shape, fill_value=default, dtype=dtype)


def _as_tuple(
    axis: int | Iterable[int] | tuple[int, ...] | None,
    ndim: int,
    default_0D_output: bool,
) -> tuple[int, ...]:
    if axis is None:
        return tuple(range(ndim)) if default_0D_output else ()

    elif not isinstance(axis, Iterable):
        return (axis,)

    return tuple(axis)


def _get_indices(
    index: tuple[SingleIndex | FullSlice, ...],
    axis: tuple[int, ...],
    where_compute: MaskWhere,
    where_output: MaskWhere,
    output_ndim: int,
) -> tuple[WHERE_SELECTION, tuple[slice, ...] | tuple[()], WHERE_SELECTION]:
    # compute on whole array at once
    if len(index) == 1 and isinstance(index[0], FullSlice) and index[0].is_whole_axis:
        return where_compute[:], (), where_output[index]

    where_to_compute = where_compute[map_slice(index)]

    # 0D output array (no chunk selection, no where selection)
    if output_ndim == 0:
        return where_to_compute, (), ()

    # nD output array
    selected_index = map_slice(tuple(e for i, e in enumerate(index) if i not in axis))
    return where_to_compute, selected_index, where_output[selected_index]


def _apply_operation(
    operation: ApplyOperation,
    dest: H5Array[Any] | npt.NDArray[Any],
    chunk_selection: tuple[slice, ...] | tuple[()],
    where_to_output: WHERE_SELECTION,
    values: npt.NDArray[Any],
) -> None:
    if values.ndim == 0:
        values = values[()]

    if operation is ApplyOperation.set:
        if dest.ndim == 0:
            dest[()] = values

        else:
            dest[chunk_selection][where_to_output] = values[where_to_output]

    elif operation is ApplyOperation.iadd:
        if dest.ndim == 0:
            dest[()] = values + dest[()]

        else:
            dest[chunk_selection][where_to_output] += values[where_to_output]

    elif operation is ApplyOperation.imul:
        if dest.ndim == 0:
            dest *= values  # type: ignore[assignment]

        else:
            dest[chunk_selection][where_to_output] *= values[where_to_output]

    elif operation is ApplyOperation.iand:
        if dest.ndim == 0:
            dest &= values  # type: ignore[assignment]

        else:
            dest[chunk_selection][where_to_output] &= values[where_to_output]

    elif operation is ApplyOperation.ior:
        if dest.ndim == 0:
            dest |= values  # type: ignore[assignment]

        else:
            dest[chunk_selection][where_to_output] |= values[where_to_output]

    else:
        raise NotImplementedError(f"Do not know how to apply operation '{operation}'")


def apply(
    func: partial[NP_FUNC],
    operation: ApplyOperation,
    a: H5Array[Any],
    out: H5Array[Any] | npt.NDArray[Any] | None,
    *,
    dtype: npt.DTypeLike | None,
    initial: int | float | complex | NoValue,
    where: npt.NDArray[np.bool_] | Iterable[np.bool_] | int | bool | NoValue,
    default_0D_output: bool = True,
) -> Any:
    dtype = a.dtype if dtype is None else dtype
    axis = _as_tuple(func.keywords.get("axis", None), a.ndim, default_0D_output)
    output_array = _get_output_array(out, a.shape, axis, func.keywords.get("keepdims", False), dtype, initial)

    if where is not False:
        where_compute = MaskWhere(where, a.shape)
        where_output = MaskWhere(where, output_array.shape)

        for index, chunk in a.iter_chunks(keepdims=True):
            where_to_compute, chunk_selection, where_to_output = _get_indices(
                index, axis, where_compute, where_output, output_array.ndim
            )
            result = np.array(
                func(chunk, where=True if where_to_compute == () else where_to_compute),
                dtype=output_array.dtype,
            )
            _apply_operation(operation, output_array, chunk_selection, where_to_output, result)

    if out is None and output_array.ndim == 0:
        return output_array[()]

    return output_array


def apply_everywhere(
    func: partial[NP_FUNC],
    operation: ApplyOperation,
    a: H5Array[Any],
    out: H5Array[Any] | npt.NDArray[Any] | None,
    *,
    dtype: npt.DTypeLike | None,
    initial: int | float | complex | NoValue,
    default_0D_output: bool = True,
) -> Any:
    dtype = a.dtype if dtype is None else dtype
    axis = _as_tuple(func.keywords.get("axis", None), a.ndim, default_0D_output)
    output_array = _get_output_array(out, a.shape, axis, func.keywords.get("keepdims", False), dtype, initial)

    where_compute = MaskWhere(True, a.shape)
    where_output = MaskWhere(True, output_array.shape)

    for index, chunk in a.iter_chunks(keepdims=True):
        where_to_compute, chunk_selection, where_to_output = _get_indices(
            index, axis, where_compute, where_output, output_array.ndim
        )
        result = np.array(func(chunk), dtype=output_array.dtype)
        _apply_operation(operation, output_array, chunk_selection, where_to_output, result)

    if out is None and output_array.ndim == 0:
        return output_array[()]

    return output_array


def _get_str_dtype(
    a: npt.NDArray[np.str_] | H5Array[np.str_],
    b: npt.NDArray[Any] | H5Array[Any],
    func: NP_FUNC,
) -> np.dtype[Any]:
    assert np.issubdtype(a.dtype, str)

    if func == np.char.multiply:
        assert np.issubdtype(b.dtype, np.number)
        return np.dtype(f"<U{a.dtype.itemsize // 4 * max(b)}")

    if func == np.char.add:
        assert np.issubdtype(b.dtype, str)
        return np.dtype(f"<U{(a.dtype.itemsize + b.dtype.itemsize) // 4}")

    elif func in (
        np.char.greater,
        np.char.greater_equal,
        np.char.less,
        np.char.less_equal,
        np.char.equal,
        np.char.not_equal,
    ):
        return np.dtype(bool)

    raise NotImplementedError


def str_apply_2(
    func: NP_FUNC,
    a: str | npt.NDArray[np.str_] | Iterable[str] | H5Array[np.str_],
    b: Any,
) -> npt.NDArray[Any] | H5Array[Any] | bool:
    if not isinstance(a, (np.ndarray, ch5mpy.H5Array)):
        a = np.array(a, dtype=str)

    if not isinstance(b, (np.ndarray, ch5mpy.H5Array)):
        b = np.array(b)

    if b.dtype == object:
        b = b.astype(str)

    output_array = _get_output_array_2(None, a.shape, b.shape, _get_str_dtype(a, b, func), None)

    if not np.issubdtype(b.dtype, str):
        try:
            return {np.char.equal: False, np.char.not_equal: True}[func]

        except KeyError:
            raise TypeError(f"'{func}' not supported between arrays with dtypes {a.dtype} and {b.dtype}.")

    for index, chunk_x1, chunk_x2 in iter_chunks_2(a, b):
        output_array[map_slice(index)] = func(chunk_x1, chunk_x2)

    return output_array


num_to_str_ufunc: dict[NP_FUNC, NP_FUNC] = {
    np.add: np.char.add,
    np.multiply: np.char.multiply,
    np.greater: np.char.greater,
    np.greater_equal: np.char.greater_equal,
    np.less: np.char.less,
    np.less_equal: np.char.less_equal,
    np.equal: np.char.equal,
    np.not_equal: np.char.not_equal,
}


def apply_2(
    func: NP_FUNC,
    a: H5Array[Any],
    b: npt.NDArray[Any] | Iterable[Any] | Number | H5Array[Any],
    *,
    out: H5Array[Any] | npt.NDArray[Any] | None,
    default: Any,
    dtype: npt.DTypeLike | None,
    where: npt.NDArray[np.bool_] | Iterable[bool] | int | bool,
) -> Any:
    # if a is a str H5Array (got here because the operators ==, +, *, ... were used), pass to str_apply_2()
    if np.issubdtype(a.dtype, str):
        assert out is None and where is True
        return str_apply_2(num_to_str_ufunc[func], a, b)

    # operation on regular H5Arrays
    if not isinstance(b, (np.ndarray, ch5mpy.H5Array)):
        b = np.array(b)

    output_array = _get_output_array_2(out, a.shape, b.shape, dtype, default)

    if where is not False:
        where_compute = MaskWhere(where, a.shape)
        where_output = MaskWhere(where, output_array.shape)

        for index, chunk_x1, chunk_x2 in iter_chunks_2(a, b):
            where_to_compute, chunk_selection, where_to_output = _get_indices(
                index, (), where_compute, where_output, output_array.ndim
            )
            result = np.array(
                func(
                    chunk_x1,
                    chunk_x2,
                    where=True if where_to_compute == () else where_to_compute,
                ),
                dtype=output_array.dtype,
            )
            _apply_operation(ApplyOperation.set, output_array, chunk_selection, where_to_output, result)

    if out is None and output_array.ndim == 0:
        return output_array[()]

    return output_array



from __future__ import annotations

from enum import Enum
from typing import Any

import numpy as np
import pytest

from ch5mpy import File, Group
from ch5mpy.write import write_dataset, write_object


class State(Enum):
    RNA = "RNA"


@pytest.mark.parametrize("obj, expected", [(1, 1), ("abc", "abc"), (None, None)])
def test_should_write_simple_attribute(obj, expected, group):
    group.attrs["something"] = obj

    assert "something" in group.attrs.keys()
    assert group.attrs["something"] == expected


def test_should_write_list_attribute(group):
    group.attrs["something"] = [1, 2, 3]

    assert "something" in group.attrs.keys()
    assert np.all(group.attrs["something"] == [1, 2, 3])


def test_should_write_complex_objects_as_strings_in_attributes(group):
    group.attrs["something"] = State.RNA

    assert "something" in group.attrs.keys()
    assert group.attrs["something"] == State.RNA


@pytest.mark.parametrize(
    "array, expected",
    [
        (np.array([1, 2, 3]), [1, 2, 3]),
        ([1, 2, 3], [1, 2, 3]),
        (np.array(["a", "b", "c"]), [b"a", b"b", b"c"]),
    ],
)
def test_should_write_array(group, array, expected):
    write_dataset(group, "something", array)

    assert "something" in group.keys()
    assert np.all(group["something"][()] == expected)


def test_should_write_dict_of_arrays(group):
    write_dataset(group, "some_dict", {"some_a": [1, 2, 3], "some_b": ["a", "b", "c"]})

    assert np.all(group["some_dict"]["some_a"][()] == [1, 2, 3])
    assert np.all(group["some_dict"]["some_b"][()] == [b"a", b"b", b"c"])


def are_equal(obj1: Any, obj2: File | Group) -> bool:
    if isinstance(obj1, (np.ndarray, list)):
        return np.array_equal(obj1, obj2[()])

    return obj1 == obj2


@pytest.mark.parametrize(
    "name, obj",
    [
        ("list_", [1, 2, 3]),
        ("array_", np.array([4, 5, 6])),
        # ('dict_', {'a': 'test', 'b': 1})
    ],
)
def test_write_object(group, name, obj):
    write_object(group, name, obj)

    assert are_equal(obj, group[name])



from __future__ import annotations

from numbers import Number
from typing import TYPE_CHECKING, Any, Generator, TypeVar, cast

import numpy as np
import numpy.typing as npt

import ch5mpy
from ch5mpy.h5array.chunks.repeated_array import RepeatedArray
from ch5mpy.h5array.chunks.utils import _as_valid_dtype
from ch5mpy.indexing import FullSlice, SingleIndex, map_slice

if TYPE_CHECKING:
    from ch5mpy import H5Array


_DT = TypeVar("_DT", bound=np.generic)
INF = np.iinfo(int).max


def get_work_array(
    shape: tuple[int, ...], slicer: tuple[FullSlice | SingleIndex, ...], dtype: np.dtype[_DT]
) -> npt.NDArray[_DT]:
    if len(slicer) == 1 and isinstance(slicer[0], FullSlice) and slicer[0].is_whole_axis:
        return np.empty(shape, dtype=object if np.issubdtype(dtype, str) else dtype)

    slicer_shape = tuple(len(s) for s in slicer if not isinstance(s, SingleIndex))
    return np.empty(slicer_shape, dtype=object if np.issubdtype(dtype, str) else dtype)


def _get_chunk_indices(chunk_size: int, shape: tuple[int, ...]) -> tuple[tuple[FullSlice | SingleIndex, ...], ...]:
    # special case of 0D arrays
    if len(shape) == 0:
        raise ValueError("0D array")

    if np.prod(shape) == 0:
        return (tuple(FullSlice.whole_axis(s) for s in shape),)

    rev_shape = tuple(reversed(shape))

    # not enough max mem
    if chunk_size <= 1:
        raise ValueError("Slicing is impossible because of insufficient allowed memory usage.")

    block_axes = int(np.argmax(~(np.cumprod(rev_shape + (np.inf,)) <= chunk_size)))
    size_block = chunk_size // np.cumprod(rev_shape)[block_axes - 1] if block_axes else min(rev_shape[0], chunk_size)

    if size_block == 0:
        block_axes = max(0, block_axes - 1)

    if block_axes == len(shape):
        # all array can be read at once
        return (tuple(FullSlice.whole_axis(s) for s in shape),)

    whole_axes = tuple(FullSlice.whole_axis(s) for s in rev_shape[:block_axes][::-1])
    iter_axis = rev_shape[block_axes]

    right_chunks = tuple(
        (FullSlice(s, min(s + size_block, iter_axis), 1, iter_axis), *whole_axes)
        for s in range(0, iter_axis, size_block)
    )

    if block_axes + 1 == len(shape):
        return right_chunks

    left_shape = shape[: -(block_axes + 1)]
    left_chunks = np.array(np.meshgrid(*map(range, left_shape))).T.reshape(-1, len(left_shape))

    return tuple(
        tuple(SingleIndex(_l, _s) for _l, _s in zip(left, shape)) + tuple(right)
        for left in left_chunks
        for right in right_chunks
    )


class ChunkIterator:
    def __init__(self, array: H5Array[Any], keepdims: bool = False):
        self._array = array
        self._keepdims = keepdims

        self._chunk_indices = _get_chunk_indices(array.chunk_size, array.shape)

        self._work_array = get_work_array(array.shape, self._chunk_indices[0], dtype=array.dtype)

    def __repr__(self) -> str:
        return f"<ChunkIterator over {self._array.shape} H5Array>"

    def __iter__(
        self,
    ) -> Generator[tuple[tuple[FullSlice | SingleIndex, ...], npt.NDArray[Any]], None, None]:
        for index in self._chunk_indices:
            work_subset = map_slice(index, shift_to_zero=True)
            self._array.read_direct(self._work_array, source_sel=map_slice(index), dest_sel=work_subset)

            # cast to str if needed
            res = _as_valid_dtype(self._work_array, self._array.dtype)[work_subset]

            # reshape to keep dimensions if needed
            if self._keepdims:
                res = res.reshape((1,) * (self._array.ndim - res.ndim) + res.shape)

            yield index, res


class PairedChunkIterator:
    def __init__(
        self,
        arr_1: H5Array[Any] | npt.NDArray[Any],
        arr_2: H5Array[Any] | npt.NDArray[Any],
        keepdims: bool = False,
    ):
        broadcasted_shape = np.broadcast_shapes(arr_1.shape, arr_2.shape)
        self._arr_1 = RepeatedArray(arr_1, broadcasted_shape)
        self._arr_2 = RepeatedArray(arr_2, broadcasted_shape)

        self._keepdims = keepdims

        chunk_size = min(
            arr_1.chunk_size if isinstance(arr_1, ch5mpy.H5Array) else INF,
            arr_2.chunk_size if isinstance(arr_2, ch5mpy.H5Array) else INF,
        )

        self._chunk_indices = _get_chunk_indices(chunk_size, shape=arr_1.shape)

        self._work_array_1 = get_work_array(broadcasted_shape, self._chunk_indices[0], dtype=arr_1.dtype)
        self._work_array_2 = get_work_array(broadcasted_shape, self._chunk_indices[0], dtype=arr_2.dtype)

    def __repr__(self) -> str:
        return f"<PairedChunkIterator over 2 {self._arr_1.shape} arrays>"

    def __iter__(
        self,
    ) -> Generator[tuple[tuple[SingleIndex | FullSlice, ...], npt.NDArray[Any], npt.NDArray[Any]], None, None]:
        for index in self._chunk_indices:
            work_subset = map_slice(index, shift_to_zero=True)
            res_1 = self._arr_1.read(self._work_array_1, map_slice(index), work_subset)
            res_2 = self._arr_2.read(self._work_array_2, map_slice(index), work_subset)

            if self._keepdims:
                res_1 = res_1.reshape((1,) * (self._arr_1.ndim - res_1.ndim) + res_1.shape)
                res_2 = res_2.reshape((1,) * (self._arr_2.ndim - res_2.ndim) + res_2.shape)

            yield index, res_1, res_2


def iter_chunks_2(
    x1: npt.NDArray[Any] | H5Array[Any], x2: npt.NDArray[Any] | H5Array[Any]
) -> Generator[tuple[tuple[SingleIndex | FullSlice, ...], npt.NDArray[Any], npt.NDArray[Any] | Number], None, None,]:
    # special case where x2 is a 0D array, iterate through chunks of x1 and always yield x2
    if x2.ndim == 0:
        if isinstance(x1, ch5mpy.H5Array):
            for chunk, arr in ChunkIterator(x1):
                yield chunk, arr, cast(Number, x2[()])

        else:
            yield (FullSlice.whole_axis(x1.shape[0]),), x1, cast(Number, x2[()])

    # nD case
    else:
        yield from PairedChunkIterator(x1, x2)
